"""Keyboard interface of picframe."""

import logging
import threading
import time
import pi3d


class InterfaceKbd:
    """Keyboard interface of picframe.
    
    This interface interacts via keyboard with the user to steer the image display.

    Attributes
    ----------
    controller : Controler 
        Controller for picframe
   

    Methods
    -------

    """

    def __init__(self, controller):
        self.__logger = logging.getLogger("interface_kbd.InterfaceKbd")
        self.__logger.info('creating an instance of InterfaceKbd')
        self.__controller = controller
        self.__keyboard = pi3d.Keyboard()
        self.__keep_looping = True
        t = threading.Thread(target=self.__loop)
        t.start()

    def __loop(self):
        while self.__keep_looping:
            key = self.__keyboard.read()
            if key == 27:
                self.__keep_looping = False
            elif key == ord('a'):
                self.__controller.back()
            elif key == ord('d'):
                self.__controller.next()
            elif key == ord('l'):
                if self.__controller.text_is_on("location"):
                    self.__controller.set_show_text("location", "OFF")
                else:
                    self.__controller.set_show_text("location", "ON")
            time.sleep(0.025)
        self.__keyboard.close() # contains references to Display instance
        self.__controller.stop()
    


import os
import json
import urllib.request
import locale
import logging
import time

URL = "https://nominatim.openstreetmap.org/reverse?format=geojson&lat={}&lon={}&zoom={}&email={}&accept-language={}"

class GeoReverse:
    def __init__(self, geo_key, zoom=18, key_list=None):
        self.__logger = logging.getLogger("geo_reverse.GeoReverse")
        self.__geo_key = geo_key
        self.__zoom = zoom
        self.__key_list = key_list
        #self.__file_path = os.path.expanduser(file_path)
        self.__geo_locations = {}
        #if os.path.isfile(file_path):
        #    with open(file_path) as f:
        #        for line in f:
        #            if line == '\n':
        #                continue
        #            (name, var) = line.partition('=')[::2]
        #            self.__geo_locations[name] = var.rstrip('\n')
        self.__language = locale.getlocale()[0][:2]

    def get_address(self, lat, lon):
        try:
            with urllib.request.urlopen(URL.format(lat, lon, self.__zoom, self.__geo_key, self.__language),
                                        timeout=3.0) as req:
                    data = json.loads(req.read().decode())
            adr = data['features'][0]['properties']['address']
            # some experimentation might be needed to get a good set of alternatives in key_list
            adr_list = []
            if self.__key_list is not None:
                for part in self.__key_list:
                    for option in part:
                        if option in adr:
                            adr_list.append(adr[option])
                            break # add just the first one from the options
            else:
                adr_list = adr.values()
            return ", ".join(adr_list)
        except Exception as e: #TODO return different thing for different exceptions
            self.__logger.error("lat=%f, lon=%f -> %s", lat, lon, e)
            return ""



import exifread
import logging
from PIL import Image

class GetImageMeta:

    def __init__(self, filename):
        self.__logger = logging.getLogger("get_image_meta.GetImageMeta")
        self.__tags = {}
        self.__filename = filename # in case no exif data in which case needed for size
        try:
            with open(filename, 'rb') as fh:
                self.__tags = exifread.process_file(fh, details=False) 
        except OSError as e:
            self.__logger.warning("Can't open file: \"%s\"", filename)
            self.__logger.warning("Cause: %s", e)
            #raise # the system should be able to withstand files being moved etc without crashing
        except Exception as e:
            self.__logger.warning("exifread doesn't manage well and gives AttributeError for heif files %s -> %s",
                                  filename, e)
        self.__do_iptc_keywords()

    def __do_iptc_keywords(self):
        try:
            from iptcinfo3 import IPTCInfo
            iptcinfo_logger = logging.getLogger('iptcinfo') # turn off useless log infos
            iptcinfo_logger.setLevel(logging.ERROR)
            with open(self.__filename, 'rb') as fh:
                iptc = IPTCInfo(fh, force=True, out_charset='utf-8') # TODO put IPTC read in separate function
                # tags
                val = iptc['keywords']
                if val is not None and len(val) > 0: 
                    keywords = ''
                    for key in iptc['keywords']:
                        keywords += key.decode('utf-8')  + ','  # decode binary strings
                    self.__tags['IPTC Keywords'] = keywords
                # caption
                val = iptc['caption/abstract']
                if val is not None and len(val) > 0:  
                    self.__tags['IPTC Caption/Abstract'] = iptc['caption/abstract'].decode('utf8')
                # title
                val = iptc['object name']
                if val is not None and len(val) > 0:  
                    self.__tags['IPTC Object Name'] = iptc['object name'].decode('utf-8')
        except Exception as e:
            self.__logger.warning("IPTC loading has failed - if you want to use this you will need to install iptcinfo3 %s -> %s",
                                  self.__filename, e)

    def has_exif(self):
        if self.__tags == {}:
            return False
        else:
            return True

    def __get_if_exist(self, key):
        if key in self.__tags:
            return self.__tags[key]
        return None

    def __convert_to_degress(self, value):
        (deg, min, sec) = value.values
        d = float(deg.num) / float(deg.den if deg.den > 0 else 1) #TODO better catching?
        m = float(min.num) / float(min.den if min.den > 0 else 1)
        s = float(sec.num) / float(sec.den if sec.den > 0 else 1)
        return d + (m / 60.0) + (s / 3600.0)

    def get_location(self):
        gps = {"latitude": None, "longitude": None}
        lat = None
        lon = None

        gps_latitude = self.__get_if_exist('GPS GPSLatitude')
        gps_latitude_ref = self.__get_if_exist('GPS GPSLatitudeRef')
        gps_longitude = self.__get_if_exist('GPS GPSLongitude')
        gps_longitude_ref = self.__get_if_exist('GPS GPSLongitudeRef')

        if gps_latitude and gps_latitude_ref and gps_longitude and gps_longitude_ref:
            lat = self.__convert_to_degress(gps_latitude)
            if len(gps_latitude_ref.values) > 0 and gps_latitude_ref.values[0] == 'S':
                # assume zero length string means N
                lat = 0 - lat
            gps["latitude"] = lat
            lon = self.__convert_to_degress(gps_longitude)
            if len(gps_longitude_ref.values) and gps_longitude_ref.values[0] == 'W':
                lon = 0 - lon
            gps["longitude"] = lon
        return gps

    def get_orientation(self):
        val = self.__get_if_exist('Image Orientation')
        if val is not None:
            return int(val.values[0])
        else:
            return 1

    def get_exif(self, key):
        val = self.__get_if_exist(key)
        if val:
            if key == 'EXIF FNumber':
                val = round(val.values[0].num / val.values[0].den, 1)
            elif key in ['IPTC Keywords',  'IPTC Caption/Abstract',  'IPTC Object Name']:
                return val
            else:
                val = val.printable
        return val

    def get_size(self):
        try: # corrupt image file might crash app
            return Image.open(self.__filename).size
        except:
            return (0, 0)


import yaml
import os
import time
import logging
import random
import json
import locale
from picframe import geo_reverse, image_cache

DEFAULT_CONFIGFILE = "~/picframe_data/config/configuration.yaml"
DEFAULT_CONFIG = {
    'viewer': {
        'blur_amount': 12,
        'blur_zoom': 1.0,
        'blur_edges': False,
        'edge_alpha': 0.5,
        'fps': 20.0,
        'background': [0.2, 0.2, 0.3, 1.0],
        'blend_type': "blend", # {"blend":0.0, "burn":1.0, "bump":2.0}

        'font_file': '~/picframe_data/data/fonts/NotoSans-Regular.ttf',
        'shader': '~/picframe_data/data/shaders/blend_new',
        'show_text_fm': '%b %d, %Y',
        'show_text_tm': 20.0,
        'show_text_sz': 40,
        'show_text': "name location",
        'text_justify': 'L',
        'fit': False,
        'auto_resize': True,
        'kenburns': False,
        'display_x': 0,
        'display_y': 0,
        'display_w': None,
        'display_h': None,
        'use_glx': False,                          # default=False. Set to True on linux with xserver running
        'test_key': 'test_value',
        'mat_images': True,
        'mat_type': None,
        'outer_mat_color': None,
        'inner_mat_color': None,
        'outer_mat_border': 75,
        'inner_mat_border': 40,
        'inner_mat_use_texture': False,
        'outer_mat_use_texture': True,
        'mat_resource_folder': '~/picframe_data/data/mat',
        'codepoints': "1234567890AÄÀÆÅÃBCÇDÈÉÊEËFGHIÏÍJKLMNÑOÓÖÔŌØPQRSTUÚÙÜVWXYZaáàãæåäbcçdeéèêëfghiíïjklmnñoóôōøöpqrsßtuúüvwxyz., _-+*()&/`´'•" # limit to 121 ie 11x11 grid_size
    },
    'model': {

        'pic_dir': '~/Pictures',
        'no_files_img': '~/picframe_data/data/no_pictures.jpg',
        'subdirectory': '',
        'recent_n': 3,
        'reshuffle_num': 1,
        'time_delay': 200.0,
        'fade_time': 10.0,
        'shuffle': True,
        'image_attr': ['PICFRAME GPS'],                          # image attributes send by MQTT, Keys are taken from exifread library, 'PICFRAME GPS' is special to retrieve GPS lon/lat
        'load_geoloc': True,
        'locale': 'en_US.utf8',
        'key_list': [['tourism','amenity','isolated_dwelling'],['suburb','village'],['city','county'],['region','state','province'],['country']],
        'geo_key': 'this_needs_to@be_changed',  # use your email address
        'db_file': '~/picframe_data/data/pictureframe.db3',
        'portrait_pairs': False,
        'deleted_pictures': '~/DeletedPictures',
        'log_level': 'WARNING',
        'use_kbd': False,
    },
    'mqtt': {
        'use_mqtt': False,                          # Set tue true, to enable mqtt
        'server': '',
        'port': 8883,
        'login': '',
        'password': '',
        'tls': '',
        'device_id': 'picframe',                                 # unique id of device. change if there is more than one picture frame
    },
    'http': {
        'use_http': False,
        'path': '~/picframe_data/html',
        'port': 9000,
        'use_ssl': False,
        'keyfile': "/path/to/key.pem",
        'certfile': "/path/to/fullchain.pem"
    }
}


class Pic: #TODO could this be done more elegantly with namedtuple

    def __init__(self, fname, last_modified, file_id, orientation=1, exif_datetime=0,
                 f_number=0, exposure_time=None, iso=0, focal_length=None,
                 make=None, model=None, lens=None, rating=None, latitude=None,
                 longitude=None, width=0, height=0, is_portrait=0, location=None, title=None,
                 caption=None, tags=None):
        self.fname = fname
        self.last_modified = last_modified
        self.file_id = file_id
        self.orientation = orientation
        self.exif_datetime = exif_datetime
        self.f_number = f_number
        self.exposure_time = exposure_time
        self.iso = iso
        self.focal_length = focal_length
        self.make = make
        self.model = model
        self.lens = lens
        self.rating = rating
        self.latitude = latitude
        self.longitude = longitude
        self.width = width
        self.height = height
        self.is_portrait = is_portrait
        self.location = location
        self.tags=tags
        self.caption=caption
        self.title=title


class Model:

    def __init__(self, configfile = DEFAULT_CONFIGFILE):
        self.__logger = logging.getLogger("model.Model")
        self.__logger.debug('creating an instance of Model')
        self.__config = DEFAULT_CONFIG
        self.__last_file_change = 0.0
        self.__required_db_schema_version = 1
        configfile = os.path.expanduser(configfile)
        self.__logger.info("Open config file: %s:",configfile)
        with open(configfile, 'r') as stream:
            try:
                conf = yaml.safe_load(stream)
                for section in ['viewer', 'model', 'mqtt', 'http']:
                    self.__config[section] = {**DEFAULT_CONFIG[section], **conf[section]}

                self.__logger.debug('config data = %s', self.__config)
            except yaml.YAMLError as exc:
                self.__logger.error("Can't parse yaml config file: %s: %s", configfile, exc)
        logging.getLogger().setLevel(self.get_model_config()['log_level']) # set root logger
        self.__file_list = [] # this is now a list of tuples i.e (file_id1,) or (file_id1, file_id2)
        self.__number_of_files = 0 # this is shortcut for len(__file_list)
        self.__reload_files = True
        self.__file_index = 0 # pointer to next position in __file_list
        self.__current_pics = (None, None) # this hold a tuple of (pic, None) or two pic objects if portrait pairs
        self.__num_run_through = 0

        model_config = self.get_model_config() # alias for brevity as used several times below
        try:
            locale.setlocale(locale.LC_TIME, model_config['locale'])
        except:
            self.__logger.error("error trying to set locale to {}".format(model_config['locale']))
        self.__pic_dir = os.path.expanduser(model_config['pic_dir'])
        self.__subdirectory = os.path.expanduser(model_config['subdirectory'])
        self.__load_geoloc = model_config['load_geoloc']
        self.__geo_reverse = geo_reverse.GeoReverse(model_config['geo_key'], key_list=self.get_model_config()['key_list'])
        self.__image_cache = image_cache.ImageCache(self.__pic_dir,
                                                    os.path.expanduser(model_config['db_file']),
                                                    self.__geo_reverse,
                                                    self.__required_db_schema_version,
                                                    model_config['portrait_pairs'])
        self.__deleted_pictures = model_config['deleted_pictures']
        self.__no_files_img = os.path.expanduser(model_config['no_files_img'])
        self.__where_clauses = {} # these will be modified by controller

    def get_viewer_config(self):
        return self.__config['viewer']

    def get_model_config(self):
        return self.__config['model']

    def get_mqtt_config(self):
        return self.__config['mqtt']

    def get_http_config(self):
        return self.__config['http']

    @property
    def fade_time(self):
        return self.__config['model']['fade_time']

    @fade_time.setter
    def fade_time(self, time):
        self.__config['model']['fade_time'] = time

    @property
    def time_delay(self):
        return self.__config['model']['time_delay']

    @time_delay.setter
    def time_delay(self, time):
        self.__config['model']['time_delay'] = time

    @property
    def subdirectory(self):
        return self.__subdirectory

    @subdirectory.setter
    def subdirectory(self, dir):
        _, root = os.path.split(self.__pic_dir)
        actual_dir = root
        if self.subdirectory != '':
            actual_dir = self.subdirectory
        if actual_dir != dir:
            if root == dir:
                self.__subdirectory = ''
            else:
                self.__subdirectory = dir
            self.__logger.info("Set subdirectory to: %s", self.__subdirectory)
            self.__reload_files = True

    @property
    def EXIF_TO_FIELD(self): # bit convoluted TODO hold in config? not really configurable
        return self.__image_cache.EXIF_TO_FIELD

    @property
    def shuffle(self):
        return self.__config['model']['shuffle']

    @shuffle.setter
    def shuffle(self, val:bool):
        self.__config['model']['shuffle'] = val #TODO should this be altered in config?
        #if val == True:
        #    self.__shuffle_files()
        #else:
        #    self.__sort_files()
        self.__reload_files = True

    def set_where_clause(self, key, value=None):
        # value must be a string for later join()
        if (value is None or len(value) == 0):
            if key in self.__where_clauses:
                self.__where_clauses.pop(key)
            return
        self.__where_clauses[key] = value

    def pause_looping(self, val):
        self.__image_cache.pause_looping(val)

    def stop_image_chache(self):
        self.__image_cache.stop()

    def get_directory_list(self):
        _, root = os.path.split(self.__pic_dir)
        actual_dir = root
        if self.subdirectory != '':
            actual_dir = self.subdirectory
        subdir_list = next(os.walk(self.__pic_dir))[1]
        subdir_list.insert(0,root)
        return actual_dir, subdir_list

    def force_reload(self):
        self.__reload_files = True

    def set_next_file_to_previous_file(self):
        self.__file_index = (self.__file_index - 2) % self.__number_of_files # TODO deleting last image results in ZeroDivisionError

    def get_next_file(self):
        if self.__reload_files:
            for i in range(5): # give image_cache chance on first load if a large directory
                self.__get_files()
                if self.__number_of_files > 0:
                    break
                time.sleep(0.5)
        if self.__file_index == self.__number_of_files:
            self.__num_run_through += 1
            if self.shuffle and self.__num_run_through >= self.get_model_config()['reshuffle_num']:
                #self.__num_run_through = 0
                #self.__shuffle_files()
                self.__reload_files = True
            self.__file_index = 0
        if self.__number_of_files == 0:
            pic = Pic(self.__no_files_img, 0, 0)
            paired_pic = None
        else:
            file_ids = self.__file_list[self.__file_index]
            pic_row = self.__image_cache.get_file_info(file_ids[0])
            pic = Pic(**pic_row) if pic_row is not None else None
            if len(file_ids) == 2:
                pic_row = self.__image_cache.get_file_info(file_ids[1])
                paired_pic = Pic(**pic_row) if pic_row is not None else None
            else:
                paired_pic = None
        self.__current_pics = (pic, paired_pic)
        self.__file_index += 1 # don't wrap back as __file_index == __number_of_files used as trigger above
        return self.__current_pics

    def get_number_of_files(self):
        return self.__number_of_files

    def get_current_pics(self):
        return self.__current_pics

    def delete_file(self):
        # delete the current pic. If it's a portrait pair then only the left one will be deleted
        pic = self.__current_pics[0]
        if pic is None:
            return None
        f_to_delete = pic.fname
        move_to_dir = os.path.expanduser(self.__deleted_pictures)
        # TODO should these os system calls be inside a try block in case the file has been deleted after it started to show?
        if not os.path.exists(move_to_dir):
          os.system("mkdir {}".format(move_to_dir)) # problems with ownership using python func
        os.system("mv '{}' '{}'".format(f_to_delete, move_to_dir)) # and with SMB drives
        # find and delete record from __file_list
        for i, file_rec in enumerate(self.__file_list):
            if file_rec[0] == pic.file_id: # database id TODO check that db tidies itself up
                self.__file_list.pop(i)
                self.__number_of_files -= 1
                break

    def __get_files(self):
        if self.subdirectory != "":
            picture_dir = os.path.join(self.__pic_dir, self.subdirectory) # TODO catch, if subdirecotry does not exist
        else:
            picture_dir = self.__pic_dir
        where_list = ["fname LIKE '{}/%'".format(picture_dir)] # TODO / on end to stop 'test' also selecting test1 test2 etc
        where_list.extend(self.__where_clauses.values())

        if len(where_list) > 0:
            where_clause = " AND ".join(where_list) # TODO now always true - remove unreachable code
        else:
            where_clause = "1"

        sort_list = []
        recent_n = self.get_model_config()["recent_n"]
        if recent_n > 0:
            sort_list.append("last_modified < {:.0f}".format(time.time() - 3600 * 24 * recent_n))

        if self.shuffle:
            sort_list.append("RANDOM()")
        else:
            sort_list.append("exif_datetime ASC")
        sort_clause = ",".join(sort_list)

        self.__file_list = self.__image_cache.query_cache(where_clause, sort_clause)
        self.__number_of_files = len(self.__file_list)
        self.__file_index = 0
        self.__num_run_through = 0
        self.__reload_files = False

    """def __shuffle_files(self):
        #self.__file_list.sort(key=lambda x: x[1]) # will be later files last
        recent_n = self.get_model_config()['recent_n']
        temp_list_first = self.__file_list[-recent_n:]
        temp_list_last = self.__file_list[:-recent_n]
        random.seed()
        random.shuffle(temp_list_first)
        random.shuffle(temp_list_last)
        self.__file_list = temp_list_first + temp_list_last

    def __sort_files(self):
        self.__file_list.sort() # if not shuffled; sort by name"""



"""MQTT interface of picframe."""

import logging
import time
import paho.mqtt.client as mqtt
import json
import os
from picframe import __version__


class InterfaceMQTT:
    """MQTT interface of picframe.
    
    This interface interacts via mqtt with the user to steer the image display.

    Attributes
    ----------
    controller : Controler 
        Controller for picframe
   

    Methods
    -------

    """

    def __init__(self, controller, mqtt_config):
        self.__logger = logging.getLogger("interface_mqtt.InterfaceMQTT")
        self.__logger.info('creating an instance of InterfaceMQTT')
        self.__controller = controller
        try:
            device_id = mqtt_config['device_id']
            self.__client = mqtt.Client(client_id = device_id, clean_session=True)
            login = mqtt_config['login']
            password = mqtt_config['password']
            self.__client.username_pw_set(login, password) 
            tls = mqtt_config['tls']
            if tls:
                self.__client.tls_set(tls)
            server = mqtt_config['server']
            port = mqtt_config['port']
            self.__client.connect(server, port, 60) 
            self.__client.will_set("homeassistant/switch/" + mqtt_config['device_id'] + "/available", "offline", qos=0, retain=True)
            self.__client.on_connect = self.on_connect
            self.__client.on_message = self.on_message
            self.__device_id = mqtt_config['device_id']
        except Exception as e:
            self.__logger.info("MQTT not set up because of: {}".format(e))
    
    def start(self):
        try:
            self.__controller.publish_state = self.publish_state
            self.__client.loop_start()
        except Exception as e:
            self.__logger.info("MQTT not started because of: {}".format(e))

    def stop(self):
        try:
            self.__controller.publish_state = None
            self.__client.loop_stop()
        except Exception as e:
            self.__logger.info("MQTT stopping failed because of: {}".format(e))


    def on_connect(self, client, userdata, flags, rc):
        if rc != 0:
            self.__logger.warning("Can't connect with mqtt broker. Reason = {0}".format(rc))   
            return 
        self.__logger.info('Connected with mqtt broker')

        sensor_topic_head = "homeassistant/sensor/" + self.__device_id
        switch_topic_head = "homeassistant/switch/" + self.__device_id

        # send last will and testament
        available_topic = switch_topic_head + "/available"
        client.publish(available_topic, "online", qos=0, retain=True)

        # state_topic for all picframe sensors
        state_topic = sensor_topic_head + "/state"

        ## sensors
        self.__setup_sensor(client, sensor_topic_head, "date_from", "mdi:calendar-arrow-left", available_topic)
        self.__setup_sensor(client, sensor_topic_head, "date_to", "mdi:calendar-arrow-right", available_topic)
        self.__setup_sensor(client, sensor_topic_head, "time_delay", "mdi:image-plus", available_topic)
        self.__setup_sensor(client, sensor_topic_head, "brightness", "mdi:brightness-6", available_topic)
        self.__setup_sensor(client, sensor_topic_head, "fade_time", "mdi:image-size-select-large", available_topic)
        self.__setup_sensor(client, sensor_topic_head, "location_filter", "mdi:map-search", available_topic)
        self.__setup_sensor(client, sensor_topic_head, "tags_filter", "mdi:image-search", available_topic)
        self.__setup_sensor(client, sensor_topic_head, "image_counter", "mdi:camera-burst", available_topic)
        self.__setup_sensor(client, sensor_topic_head, "image", "mdi:file-image", available_topic, has_attributes=True)
        self.__setup_sensor(client, sensor_topic_head, "directory", "mdi:folder-multiple-image", available_topic, has_attributes=True)

        ## switches
        self.__setup_switch(client, switch_topic_head, "_text_refresh", "mdi:refresh", available_topic)
        self.__setup_switch(client, switch_topic_head, "_delete", "mdi:delete", available_topic)
        self.__setup_switch(client, switch_topic_head, "_name_toggle", "mdi:subtitles", available_topic,
                            self.__controller.text_is_on("name"))
        self.__setup_switch(client, switch_topic_head, "_title_toggle", "mdi:subtitles", available_topic,
                            self.__controller.text_is_on("title"))
        self.__setup_switch(client, switch_topic_head, "_caption_toggle", "mdi:subtitles", available_topic,
                            self.__controller.text_is_on("caption"))
        self.__setup_switch(client, switch_topic_head, "_date_toggle", "mdi:calendar-today", available_topic,
                            self.__controller.text_is_on("date"))
        self.__setup_switch(client, switch_topic_head, "_location_toggle", "mdi:crosshairs-gps", available_topic,
                            self.__controller.text_is_on("location"))
        self.__setup_switch(client, switch_topic_head, "_directory_toggle", "mdi:folder", available_topic,
                            self.__controller.text_is_on("directory"))
        self.__setup_switch(client, switch_topic_head, "_text_off", "mdi:badge-account-horizontal-outline", available_topic)
        self.__setup_switch(client, switch_topic_head, "_display", "mdi:panorama", available_topic,
                            self.__controller.display_is_on)
        self.__setup_switch(client, switch_topic_head, "_shuffle", "mdi:shuffle-variant", available_topic,
                            self.__controller.shuffle)
        self.__setup_switch(client, switch_topic_head, "_paused", "mdi:pause", available_topic,
                            self.__controller.paused)
        self.__setup_switch(client, switch_topic_head, "_back", "mdi:skip-previous", available_topic)
        self.__setup_switch(client, switch_topic_head, "_next", "mdi:skip-next", available_topic)

        client.subscribe(self.__device_id + "/stop", qos=0) # close down without killing!

    def __setup_sensor(self, client, sensor_topic_head, topic, icon, available_topic, has_attributes=False):
        config_topic = sensor_topic_head + "_" + topic + "/config"
        name = self.__device_id + "_" + topic
        if has_attributes == True:
            config_payload = json.dumps({"name": name,
                                     "icon": icon,
                                     "state_topic": sensor_topic_head + "/state",
                                     "value_template": "{{ value_json." + topic + "}}",
                                     "avty_t": available_topic,
                                     "json_attributes_topic": sensor_topic_head + "_" + topic + "/attributes",
                                     "uniq_id": name,
                                     "dev":{"ids":[self.__device_id]}})
        else:
            config_payload = json.dumps({"name": name,
                                     "icon": icon,
                                     "state_topic": sensor_topic_head + "/state",
                                     "value_template": "{{ value_json." + topic + "}}",
                                     "avty_t": available_topic,
                                     "uniq_id": name,
                                     "dev":{"ids":[self.__device_id]}})
        client.publish(config_topic, config_payload, qos=0, retain=True)
        client.subscribe(self.__device_id + "/" + topic, qos=0)

    def __setup_switch(self, client, switch_topic_head, topic, icon,
                       available_topic, is_on=False):
        config_topic = switch_topic_head + topic + "/config"
        command_topic = switch_topic_head + topic + "/set"
        state_topic = switch_topic_head + topic + "/state"
        config_payload = json.dumps({"name": self.__device_id + topic,
                                     "icon": icon,
                                     "command_topic": command_topic,
                                     "state_topic": state_topic,
                                     "avty_t": available_topic,
                                     "uniq_id": self.__device_id + topic,
                                     "dev": {
                                        "ids": [self.__device_id], 
                                        "name": self.__device_id, 
                                        "mdl": "PictureFrame", 
                                        "sw": __version__, 
                                        "mf": "pi3d PictureFrame project"}})
      
        client.subscribe(command_topic , qos=0)
        client.publish(config_topic, config_payload, qos=0, retain=True)
        client.publish(state_topic, "ON" if is_on else "OFF", qos=0, retain=True)

    def on_message(self, client, userdata, message):
        msg = message.payload.decode("utf-8") 
        switch_topic_head = "homeassistant/switch/" + self.__device_id
       
        ###### switches ######
        # display
        if message.topic == switch_topic_head + "_display/set":
            state_topic = switch_topic_head + "_display/state"
            if msg == "ON":
                self.__controller.display_is_on = True
                client.publish(state_topic, "ON", retain=True)
            elif msg == "OFF":
                self.__controller.display_is_on = False
                client.publish(state_topic, "OFF", retain=True)
        # shuffle
        elif message.topic == switch_topic_head + "_shuffle/set":
            state_topic = switch_topic_head + "_shuffle/state"
            if msg == "ON":
                self.__controller.shuffle = True
                client.publish(state_topic, "ON", retain=True)
            elif msg == "OFF":
                self.__controller.shuffle = False
                client.publish(state_topic, "OFF", retain=True)
        # paused
        elif message.topic == switch_topic_head + "_paused/set":
            state_topic = switch_topic_head + "_paused/state"
            if msg == "ON":
                self.__controller.paused = True
                client.publish(state_topic, "ON", retain=True)
            elif msg == "OFF":
                self.__controller.paused = False
                client.publish(state_topic, "OFF", retain=True)
        # back buttons
        elif message.topic == switch_topic_head + "_back/set":
            state_topic = switch_topic_head + "_back/state"
            if msg == "ON":
                client.publish(state_topic, "OFF", retain=True)
                self.__controller.back()
        # next buttons
        elif message.topic == switch_topic_head + "_next/set":
            state_topic = switch_topic_head + "_next/state"
            if msg == "ON":
                client.publish(state_topic, "OFF", retain=True)
                self.__controller.next()
        # delete
        elif message.topic == switch_topic_head + "_delete/set":
            state_topic = switch_topic_head + "_delete/state"
            if msg == "ON":
                client.publish(state_topic, "OFF", retain=True)
                self.__controller.delete()
        # title on
        elif message.topic == switch_topic_head + "_title_toggle/set":
            state_topic = switch_topic_head + "_title_toggle/state"
            if msg in ("ON", "OFF"):
                self.__controller.set_show_text("title", msg)
                client.publish(state_topic, msg, retain=True)
        # caption on
        elif message.topic == switch_topic_head + "_caption_toggle/set":
            state_topic = switch_topic_head + "_caption_toggle/state"
            if msg in ("ON", "OFF"):
                self.__controller.set_show_text("caption", msg)
                client.publish(state_topic, msg, retain=True)
        # name on
        elif message.topic == switch_topic_head + "_name_toggle/set":
            state_topic = switch_topic_head + "_name_toggle/state"
            if msg in ("ON", "OFF"):
                self.__controller.set_show_text("name", msg)
                client.publish(state_topic, msg, retain=True)
        # date_on
        elif message.topic == switch_topic_head + "_date_toggle/set":
            state_topic = switch_topic_head + "_date_toggle/state"
            if msg in ("ON", "OFF"):
                self.__controller.set_show_text("date", msg)
                client.publish(state_topic, msg, retain=True)
        # location_on
        elif message.topic == switch_topic_head + "_location_toggle/set":
            state_topic = switch_topic_head + "_location_toggle/state"
            if msg in ("ON", "OFF"):
                self.__controller.set_show_text("location", msg)
                client.publish(state_topic, msg, retain=True)
        # directory_on
        elif message.topic == switch_topic_head + "_directory_toggle/set":
            state_topic = switch_topic_head + "_directory_toggle/state"
            if msg in ("ON", "OFF"):
                self.__controller.set_show_text("directory", msg)
                client.publish(state_topic, msg, retain=True)
        # text_off
        elif message.topic == switch_topic_head + "_text_off/set":
            state_topic = switch_topic_head + "_text_off/state"
            if msg == "ON":
                self.__controller.set_show_text()
                client.publish(state_topic, "OFF", retain=True)
                state_topic = switch_topic_head + "_directory_toggle/state"
                client.publish(state_topic, "OFF", retain=True)
                state_topic = switch_topic_head + "_location_toggle/state"
                client.publish(state_topic, "OFF", retain=True)
                state_topic = switch_topic_head + "_date_toggle/state"
                client.publish(state_topic, "OFF", retain=True)
                state_topic = switch_topic_head + "_name_toggle/state"
                client.publish(state_topic, "OFF", retain=True)
                state_topic = switch_topic_head + "_title_toggle/state"
                client.publish(state_topic, "OFF", retain=True)
                state_topic = switch_topic_head + "_caption_toggle/state"
                client.publish(state_topic, "OFF", retain=True)
        # text_refresh
        elif message.topic == switch_topic_head + "_text_refresh/set":
            state_topic = switch_topic_head + "_text_refresh/state"
            if msg == "ON":
                client.publish(state_topic, "OFF", retain=True)
                self.__controller.refresh_show_text()

        ##### values ########
        # change subdirectory
        elif message.topic == self.__device_id + "/directory":
            self.__logger.info("Recieved subdirectory: %s", msg)
            self.__controller.subdirectory = msg
        # date_from
        elif message.topic == self.__device_id + "/date_from":
            self.__logger.info("Recieved date_from: %s", msg)
            self.__controller.date_from = msg
        # date_to
        elif message.topic == self.__device_id + "/date_to":
            self.__logger.info("Recieved date_to: %s", msg)
            self.__controller.date_to = msg
        # fade_time
        elif message.topic == self.__device_id + "/fade_time":
            self.__logger.info("Recieved fade_time: %s", msg)
            self.__controller.fade_time = float(msg)
        # time_delay
        elif message.topic == self.__device_id + "/time_delay":
            self.__logger.info("Recieved time_delay: %s", msg)
            self.__controller.time_delay = float(msg)
        # brightness
        elif message.topic == self.__device_id + "/brightness":
            self.__logger.info("Recieved brightness: %s", msg)
            self.__controller.brightness = float(msg)
        # location filter
        elif message.topic == self.__device_id + "/location_filter":
            self.__logger.info("Recieved location filter: %s", msg)
            self.__controller.location_filter = msg
        # tags filter
        elif message.topic == self.__device_id + "/tags_filter":
            self.__logger.info("Recieved tags filter: %s", msg)
            self.__controller.tags_filter = msg

        # stop loops and end program
        elif message.topic == self.__device_id + "/stop":
            self.__controller.stop()

    def publish_state(self, image, image_attr):
        topic_head =  "homeassistant/sensor/" + self.__device_id
        switch_topic_head = "homeassistant/switch/" + self.__device_id
        state_topic = topic_head + "/state"
        state_payload = {}
        # directory sensor
        actual_dir, dir_list = self.__controller.get_directory_list()
        state_payload["directory"] = actual_dir
        dir_attr = {}
        dir_attr['directories'] = dir_list
        # image counter sensor
        state_payload["image_counter"] = str(self.__controller.get_number_of_files()) 
        # image sensor
        _, tail = os.path.split(image)
        state_payload["image"] = tail
        # date_from
        state_payload["date_from"] = int(self.__controller.date_from)
        # date_to
        state_payload["date_to"] = int(self.__controller.date_to)
        # time_delay
        state_payload["time_delay"] = self.__controller.time_delay
        # fade_time
        state_payload["fade_time"] = self.__controller.fade_time
        # brightness
        state_payload["brightness"] = self.__controller.brightness
        # location_filter
        state_payload["location_filter"] = self.__controller.location_filter
        # tags_filter
        state_payload["tags_filter"] = self.__controller.tags_filter

        # send last will and testament
        available_topic = switch_topic_head + "/available"
        self.__client.publish(available_topic, "online", qos=0, retain=True)

        #pulish sensors
        attributes_topic = topic_head + "_image/attributes"
        self.__logger.debug("Send image attributes: %s", image_attr)
        self.__client.publish(attributes_topic, json.dumps(image_attr), qos=0, retain=False)
        attributes_topic = topic_head + "_directory/attributes"
        self.__client.publish(attributes_topic, json.dumps(dir_attr), qos=0, retain=False)
        self.__logger.info("Send state: %s", state_payload)
        self.__client.publish(state_topic, json.dumps(state_payload), qos=0, retain=False)



import pytest
import logging


from  picframe.get_image_meta import GetImageMeta

logger = logging.getLogger("test_get_image_data")
logger.setLevel(logging.DEBUG)

def test_file_not_found():
    with pytest.raises(OSError):
        GetImageMeta("nonsense")

def test_open_file():
    try:
        GetImageMeta("test/AlleExif.JPG")
    except:
        pytest.fail("Unexpected exception")

def test_has_exif():
    try:
        exifs = GetImageMeta("test/AlleExif.JPG")
        assert exifs.has_exif() == True
    except:
        pytest.fail("Unexpected exception")

def test_has_no_exif():
    try:
        exifs = GetImageMeta("test/noimage.jpg")
        assert exifs.has_exif() == False
    except:
        pytest.fail("Unexpected exception")

def test_get_location():
    try:
        exifs = GetImageMeta("test/AlleExif.JPG")
        gps = exifs.get_locaction()
        assert  gps == {"latitude": 25.197269, "longitude": 55.274359}
    except:
        pytest.fail("Unexpected exception")

def test_get_no_location():
    try:
        exifs = GetImageMeta("test/noimage.jpg")
        gps = exifs.get_locaction()
        assert  gps == {"latitude": None, "longitude": None}
    except:
        pytest.fail("Unexpected exception")

def test_get_exif():
    try:
        exifs = GetImageMeta("test/AlleExif.JPG")
        val = exifs.get_exif('EXIF FNumber')
        assert  val == {"EXIF FNumber": 2.8}
        val = exifs.get_exif('EXIF ExposureTime')
        assert  val == {"EXIF ExposureTime": "1/30"}
        val = exifs.get_exif('EXIF ISOSpeedRatings')
        assert  val == {"EXIF ISOSpeedRatings": "6400"}
        val = exifs.get_exif('EXIF FocalLength')
        assert  val == {"EXIF FocalLength": "17"}
        val = exifs.get_exif('EXIF DateTimeOriginal')
        assert  val == {"EXIF DateTimeOriginal": "2020:01:30 20:01:28"}
        val = exifs.get_exif('Image Model')
        assert  val == {"Image Model": "ILCE-7RM3"}
    except:
        pytest.fail("Unexpected exception")

def test_get_orientation():
    try:
        exifs = GetImageMeta("test/noimage.jpg")
        orientation = exifs.get_orientation()
        assert  orientation == 1
    except:
        pytest.fail("Unexpected exception")


from PIL import Image, ImageOps, ImageDraw
from ninepatch import Ninepatch
import numpy as np
import random
import logging
import time

class MatImage:

    # region Constructor

    def __init__(self, display_size, mat_type = None, outer_mat_color = None,
                resource_folder='.', inner_mat_color = None, outer_mat_border = 75,
                inner_mat_border = 40, outer_mat_use_texture = True,
                inner_mat_use_texture = False, auto_inner_mat_color = True):

        self.__mat_types = ['float', 'float_polaroid', 'float_color_wrap', 'single_bevel', 'double_bevel', 'double_flat']

        self.__logger = logging.getLogger("mat_image.MatImage")

        self.auto_inner_mat_color = auto_inner_mat_color
        self.display_size = display_size
        self.inner_mat_border = inner_mat_border
        self.inner_mat_color = inner_mat_color
        self.mat_type = mat_type
        self.outer_mat_border = outer_mat_border
        self.outer_mat_color = outer_mat_color
        self.outer_mat_use_texture = outer_mat_use_texture
        self.inner_mat_use_texture = inner_mat_use_texture

        # --- Matting resources ---
        self.__mat_texture = Image.open('{0}/mat_texture.jpg'.format(resource_folder)).convert("L")
        self.__9patch_bevel = Ninepatch('{0}/9_patch_bevel.png'.format(resource_folder))
        self.__9patch_drop_shadow = Ninepatch('{0}/9_patch_drop_shadow.png'.format(resource_folder))
        self.__9patch_inner_shadow = Ninepatch('{0}/9_patch_inner_shadow.png'.format(resource_folder))
        self.__9patch_highlight = Ninepatch('{0}/9_patch_highlight.png'.format(resource_folder))

    # endregion Constructor

    # region Public Properties

    @property
    def display_size(self):
        return self.__display_size

    @display_size.setter
    def display_size(self, val):
        self.__display_size = val
        self.__display_width, self.__display_height = val

    @property
    def display_width(self):
        return self.__display_width

    @property
    def display_height(self):
        return self.__display_height

    @property
    def outer_mat_border(self):
        return self.__outer_mat_border

    @outer_mat_border.setter
    def outer_mat_border(self, val):
        self.__outer_mat_border = val

    @property
    def inner_mat_border(self):
        return self.__inner_mat_border

    @inner_mat_border.setter
    def inner_mat_border(self, val):
        self.__inner_mat_border = val

    @property
    def outer_mat_color(self):
        return self.__outer_mat_color

    @outer_mat_color.setter
    def outer_mat_color(self, val):
        self.__outer_mat_color = val

    @property
    def inner_mat_color(self):
        return self.__inner_mat_color

    @inner_mat_color.setter
    def inner_mat_color(self, val):
        self.__inner_mat_color = val

    @property
    def mat_type(self):
        return self.__mat_type

    @mat_type.setter
    def mat_type(self, val):
        self.__mat_type = self.__get_mat_type_from_user_string(val)

    @property
    def mat_types(self):
        return self.__mat_types

    @property
    def outer_mat_use_texture(self):
        return self.__outer_mat_use_texture

    @outer_mat_use_texture.setter
    def outer_mat_use_texture(self, val):
        self.__outer_mat_use_texture = val

    @property
    def inner_mat_use_texture(self):
        return self.__inner_mat_use_texture

    @inner_mat_use_texture.setter
    def inner_mat_use_texture(self, val):
        self.__inner_mat_use_texture = val

    # endregion Pubic Properties

    # region Public Methods

    def mat_image(self, images):

        # Randomly pick a mat type from those specified by the User
        mat_type = random.choice(self.mat_type)

        # If a mat color wasn't specified, get one
        if not self.outer_mat_color:
            self.__outer_mat_color_save = self.__get_outer_mat_color(images[0])
        else:
            self.__outer_mat_color_save = tuple(self.outer_mat_color)

        if mat_type == 'float':
            image = self.__style_float(images)
        elif mat_type == 'float_polaroid':
            image = self.__style_float_polaroid(images)
        elif mat_type == 'float_color_wrap':
            image = self.__style_float_color_wrap(images)
        elif mat_type == 'single_bevel':
            image = self.__style_single_mat_bevel(images)
        elif mat_type == 'double_bevel':
            image = self.__style_double_mat_bevel(images)
        elif mat_type == 'double_flat':
            image = self.__style_double_mat_flat(images)
        else: image = None

        return image

    # endregion Public Methods

    # region Matting Styles

    def __style_float(self, images):
        pic_count = len(images)
        pic_wid = (self.display_width / pic_count) - (((pic_count + 1) / pic_count) * self.outer_mat_border)
        pic_height = self.display_height - (self.outer_mat_border * 2)

        final_images = []
        for image in images:
            image = self.__scale_image(image, (pic_wid, pic_height))
            self.__add_image_outline(image, self.__outer_mat_color_save, auto_adjust=True)
            image = self.__add_drop_shadow(image)
            final_images.append(image)

        return self.__layout_images(final_images)

    def __style_float_polaroid(self, images):
        border_width = 18
        pic_count = len(images)
        pic_wid = (self.display_width / pic_count) - (((pic_count + 1) / pic_count) * self.outer_mat_border) - (border_width * 2)
        pic_height = self.display_height - (self.outer_mat_border * 2) - (border_width * 2)

        final_images = []
        for image in images:
            image = self.__scale_image(image, (pic_wid, pic_height))
            self.__add_image_outline(image, self.__outer_mat_color_save)
            image = ImageOps.expand(image, border_width)
            self.__add_image_outline(image, (210,210,210), outline_width=border_width)
            image = self.__add_drop_shadow(image)
            final_images.append(image)

        return self.__layout_images(final_images)

    def __style_float_color_wrap(self, images):
        border_width = 18
        pic_count = len(images)
        pic_wid = (self.display_width / pic_count) - (((pic_count + 1) / pic_count) * self.outer_mat_border) - (border_width * 2)
        pic_height = self.display_height - (self.outer_mat_border * 2) - (border_width * 2)

        final_images = []
        for image in images:
            color = self.__get_darker_shade(self.__outer_mat_color_save, 0.35)
            color2 = self.__get_darker_shade(self.__outer_mat_color_save, 0.2)
            image = self.__scale_image(image, (pic_wid, pic_height))
            self.__add_image_outline(image, color2)
            image = ImageOps.expand(image, border_width)
            self.__add_image_outline(image, color, outline_width=border_width)
            highlight = self.__9patch_highlight.render(image.width, image.height)
            image.paste(highlight, (0,0), highlight)
            image = self.__add_drop_shadow(image)
            final_images.append(image)

        return self.__layout_images(final_images)


    def __style_single_mat_bevel(self, images):
        bevel_wid = 5
        pic_count = len(images)
        pic_wid = (self.display_width / pic_count) - (((pic_count + 1) / pic_count) * self.outer_mat_border) - (bevel_wid * 2)
        pic_height = self.display_height - (self.outer_mat_border * 2) - (bevel_wid * 2)

        final_images = []
        for image in images:
            image = self.__scale_image(image, (pic_wid, pic_height))
            image = self.__add_outer_bevel(image)
            final_images.append(image)

        return self.__layout_images(final_images)


    def __style_double_mat_bevel(self, images):
        bevel_wid = 5
        pic_count = len(images)
        pic_wid = (self.display_width / pic_count) - (((pic_count + 1) / pic_count) * self.outer_mat_border) - (self.inner_mat_border * 2) - (bevel_wid * 4)
        pic_height = self.display_height - (self.outer_mat_border * 2) - (self.inner_mat_border * 2) - (bevel_wid * 4)

        final_images = []
        for image in images:
            image = self.__scale_image(image, (pic_wid, pic_height))
            mat_size = (image.width + (self.inner_mat_border * 2) + (bevel_wid * 2), image.height + (self.inner_mat_border * 2) + (bevel_wid * 2))
            mat_image = self.__get_inner_mat(mat_size)
            mat_image = self.__add_outer_bevel(mat_image)
            image = self.__add_outer_bevel(image)
            mat_image.paste(image, (self.inner_mat_border + bevel_wid, self.inner_mat_border + bevel_wid))
            final_images.append(mat_image)

        return self.__layout_images(final_images)


    def __style_double_mat_flat(self, images):
        pic_count = len(images)
        pic_wid = (self.display_width / pic_count) - (((pic_count + 1) / pic_count) * self.outer_mat_border) - (self.inner_mat_border * 2)
        pic_height = self.display_height - (self.outer_mat_border * 2) - (self.inner_mat_border * 2)

        final_images = []
        for image in images:
            image = self.__scale_image(image, (pic_wid, pic_height))
            self.__add_image_outline(image, self.__outer_mat_color_save)
            mat_size = (image.width + (self.inner_mat_border * 2), image.height + (self.inner_mat_border * 2))
            mat_image = self.__get_inner_mat(mat_size)
            mat_image = self.__add_inner_shadow(mat_image)
            mat_image.paste(image, (self.inner_mat_border, self.inner_mat_border))
            final_images.append(mat_image)

        return self.__layout_images(final_images)

    # endregion Matting styles

    # region Helper Methods

    def __get_mat_type_from_user_string(self, mat_type_string):
        if mat_type_string == None: mat_type_string = ''

        final = []
        mat_type_string = mat_type_string.replace(',', "") # remove commas from the string
        for type in mat_type_string.split():
            if type in self.mat_types:
                final.append(type)
            else:
                self.__logger.debug('Skipping invalid mat type: %s', type)

        if not final:
            self.__logger.debug('No valid mat types defined - using: %s', self.mat_types)
            final = self.mat_types

        return final

    def __scale_image(self, image, size=None):
        if size == None:
            width, height = self.display_size
        else:
            width, height = size

        scale = min(width/image.width, height/image.height)
        image = image.resize((int(image.width * scale), int(image.height * scale)), resample=Image.BICUBIC)
        return image


    def __get_outer_mat_color(self, image):
        k = KmeansNp(k=3, max_iterations=10, size=100)
        colors = k.run(image)
        return tuple(colors[0])


    """def __get_least_gray_color(self, colors):
        dist = -1
        color = colors[0]
        for this_color in colors:
            this_dist =  max(this_color) - min(this_color)
            if this_dist > dist:
                dist = this_dist
                color = this_color
        return tuple(map(int, color))"""


    def __get_darker_shade(self, rgb_color, fractional_percent = 0.5):
        return tuple(map(lambda c: int(c * fractional_percent), rgb_color))


    def __get_colorized_mat(self, color, use_texture):
        if use_texture:
            mat_img = self.__mat_texture.copy()
            mat_img = mat_img.resize(self.display_size, resample=Image.BICUBIC)
            mat_img = ImageOps.colorize(mat_img, black="black", white=color)
        else:
            mat_img = Image.new('RGB', self.display_size, color)

        return mat_img


    def __get_inner_mat(self, size):
        w,h = size

        # If the color wasn't specified, get one
        if not self.inner_mat_color:
            color = self.__get_darker_shade(self.__outer_mat_color_save, 0.50)
        else:
            color = tuple(self.inner_mat_color)

        mat = self.__get_colorized_mat(color, self.inner_mat_use_texture)
        mat = mat.crop((0, 0, w, h))

        return mat


    def __add_outer_bevel(self, image, expand = True):
        if expand:
            image = ImageOps.expand(image, 5)
        outer_bevel_image = self.__9patch_bevel.render(image.width, image.height)
        image.paste(outer_bevel_image, (0,0), outer_bevel_image)
        return image


    def __add_inner_shadow(self, image):
        inner_shadow_image = self.__9patch_inner_shadow.render(image.width, image.height)
        image.paste(inner_shadow_image, (0,0), inner_shadow_image)
        return image


    def __add_image_outline(self, img, mat_base_color, outline_width=1, auto_adjust = False):
        if auto_adjust:
            # Calculate the outline color from the mat_color
            brightness = sum(mat_base_color[0:3]) / 3
            outline_color_offset = 30 if brightness < 127 else -30
            outline_color = tuple(map(lambda x: x + outline_color_offset, mat_base_color))
        else:
            outline_color = mat_base_color

        rect = ImageDraw.Draw(img)
        shape = [0, 0, img.width-1, img.height-1]
        rect.rectangle(shape, outline=outline_color, width=outline_width)


    def __add_drop_shadow(self, image):
        shadow_offset = 15
        mod_image = Image.new('RGBA', (image.width + shadow_offset, image.height + shadow_offset), (0,0,0,0))
        shadow_image = self.__9patch_drop_shadow.render(mod_image.width, mod_image.height)
        mod_image.paste(shadow_image, (0,0), shadow_image)
        mod_image.paste(image, (0,0))
        return mod_image


    def __layout_images(self, images):
        mat_image = self.__get_colorized_mat(self.__outer_mat_color_save, self.outer_mat_use_texture)
        total_wid = self.outer_mat_border * (len(images) + 1)
        for image in images:
            total_wid += image.width

        xloc = int((mat_image.width - total_wid) / 2)
        for image in images:
            xloc += self.outer_mat_border
            yloc =  int((mat_image.height -  image.height) / 2)
            if image.mode == 'RGBA':
                mat_image.paste(image, (xloc, yloc), image)
            else:
                mat_image.paste(image, (xloc, yloc))
            xloc += image.width

        return mat_image

    # endregion Helper functions

# region Automatic Color Selection ----

"""class Cluster(object):

    def __init__(self):
        self.pixels = []
        self.centroid = None

    def addPoint(self, pixel):
        self.pixels.append(pixel)

    def setNewCentroid(self):
        R = [colour[0] for colour in self.pixels]
        G = [colour[1] for colour in self.pixels]
        B = [colour[2] for colour in self.pixels]

        R = sum(R) / len(R)
        G = sum(G) / len(G)
        B = sum(B) / len(B)

        self.centroid = (R, G, B)
        self.pixels = []

        return self.centroid


class Kmeans(object):

    def __init__(self, k=3, max_iterations=5, min_distance=5.0, size=200):
        self.k = k
        self.max_iterations = max_iterations
        self.min_distance = min_distance
        self.size = (size, size)

    def run(self, image):
        image = image.copy()
        image.thumbnail(self.size)
        if image.mode != 'RGB':
            image = image.convert('RGB') # JAG, some numpy manipulations here don't expect an Alpha channel
        self.image = image

        self.pixels = np.array(image.getdata(), dtype=np.float)

        self.clusters = [None for i in range(self.k)]
        self.oldClusters = None

        randomPixels = random.sample(list(self.pixels), self.k)

        for idx in range(self.k):
            self.clusters[idx] = Cluster()
            self.clusters[idx].centroid = randomPixels[idx]

        iterations = 0
        self.start_clusters = [c.centroid for c in self.clusters] # make copy

        while self.shouldExit(iterations) is False:

            self.oldClusters = [cluster.centroid for cluster in self.clusters]

            for pixel in self.pixels:
                self.assignClusters(pixel)

            for cluster in self.clusters:
                if not cluster.pixels: continue
                cluster.setNewCentroid()

            iterations += 1

        return [cluster.centroid for cluster in self.clusters]

    def assignClusters(self, pixel):
        shortest = float('Inf')
        for cluster in self.clusters:
            distance = self.calcDistance(cluster.centroid, pixel)
            if distance < shortest:
                shortest = distance
                nearest = cluster

        nearest.addPoint(pixel)

    def calcDistance(self, a, b):
        result = np.sqrt(sum((a - b) ** 2))
        return result

    def shouldExit(self, iterations):

        if self.oldClusters is None:
            return False

        for idx in range(self.k):
            dist = self.calcDistance(
                np.array(self.clusters[idx].centroid),
                np.array(self.oldClusters[idx])
            )
            if dist < self.min_distance:
                return True

        if iterations <= self.max_iterations:
            return False

        return True"""

class KmeansNp:
    def __init__(self, k=3, max_iterations=5, min_distance=5.0, size=200):
        self.k = k
        self.max_iterations = max_iterations
        self.min_distance = min_distance
        self.size = (size, size)

    def run(self, image, start_clusters=None):
        image = image.copy()
        image.thumbnail(self.size)
        im = np.array(image, dtype=np.float)[:,:,:3]
        # following section can be used to give the clusters location as well as colour proximity
        #(ix0, ix1) = np.indices(im.shape[:2]) # vert,horiz pixel locations
        #ix0.shape = ix0.shape + (1,) # make same dim as im
        #ix1.shape = ix1.shape + (1,) # make same dim as im
        #im = np.append(im, ix0, axis=2) # TODO multiply by scale rel to rgb scale
        #im = np.append(im, ix1, axis=2) # im now r,g,b,u,v
        d = im.shape[-1] # 3 or 5 if u,v added
        im = im.reshape(-1, d) #NB need to use floats to avoid coercing to uint8 scrambling subtractions
        n = len(im)
        if start_clusters is None:
            centroids = im[np.random.choice(np.arange(n), self.k)]
        else:
            centroids = np.array(start_clusters, dtype=np.float)
        old_centroids = centroids.copy()
        for i in range(self.max_iterations):
            im.shape = (1, n, d) # add dimension to allow broadcasting
            centroids.shape = (self.k, 1, d) # ditto
            dists = (((im - centroids) ** 2).sum(axis=2)) ** 0.5 # euclidean distance - manhattan might be fine and faster
            ix = np.argmin(dists, axis=0) # indices of nearest centroid for each pixel
            im.shape = (n, d) # reduce dimensions for mean
            centroids.shape = (self.k, d) # ditto
            counts = np.unique(ix, return_counts=True)[1] # count the number of each index
            to_keep = [] # discard any centroids with no pixels nearest to them
            for j in range(self.k): # write back average location of all nearest pixels
                j_pixels = im[ix == j] # view into im where ix points to centroid j
                if len(j_pixels) > 0: # error if try to get mean zero length array TODO remove groups with few pixels?
                    centroids[j] = j_pixels.mean(axis=0)
                    to_keep.append(j)
            if len(to_keep) < len(centroids): # this will be relatively rare
                for j in to_keep[::-1]: # delete in reverse order of index
                    centroids = np.delete(centroids, j, axis=0)
                    old_centroids = np.delete(old_centroids, j, axis=0)
            movement = ((((centroids - old_centroids) ** 2).sum(axis=1)) ** 0.5).max()
            if movement < self.min_distance:
                break
            old_centroids = centroids.copy()

        c_max, c_min = centroids[:,:3].max(axis=1), centroids[:,:3].min(axis=1) # max, min for each centroid
        #c_lum = 0.5 * (c_max + c_min)
        #c_sat = (c_max - c_min) / (255.0 - np.abs(c_lum * 2.0 - 255.0)) # should check for lum == 255
        c_sat = c_max - c_min # value used previously includes element of lum TODO bias more to lighter using (1.5 * c_max - c_min)
        ix_order = np.argsort(c_sat)[::-1] # indices to sorted values - reversed
        return centroids[ix_order, :3].astype(np.uint8)

if __name__ == "__main__":

    save_folder = '/home/pi/pic_save'
    file1 = '/home/pi/Pictures/Sagelight/2011-01-22_12-05-18-10_edited.jpg'
    file2 = '/home/pi/Pictures/Sagelight/2011-01-22_12-06-07-10_edited.jpg'
    image1 = Image.open(file1)
    image2 = Image.open(file2)
    images = (image1, image2)

    matter = MatImage((1920, 1080))

    for mat_type in matter.mat_types:
        matter.mat_type = mat_type
        img = matter.mat_image(images)
        img.save('{0}/{1}_texture.jpg'.format(save_folder, mat_type))


"""Controller of picframe."""

import logging
import time
import json
import os
import signal
import sys

def make_date(txt):
    dt = txt.replace('/',':').replace('-',':').replace(',',':').replace('.',':').split(':')
    dt_tuple = tuple(int(i) for i in dt) #TODO catch badly formed dates?
    return time.mktime(dt_tuple + (0, 0, 0, 0, 0, 0))

class Controller:
    """Controller of picframe.
    
    This controller interacts via mqtt with the user to steer the image display.

    Attributes
    ----------
    model : Model 
        model of picframe containing config and business logic
    viewer : ViewerDisplay
        viewer of picframe representing the display
   

    Methods
    -------
    paused
        Getter and setter for pausing image display.
    next
        Show next image.
    back
        Show previous image.

    """

    def __init__(self, model, viewer):
        self.__logger = logging.getLogger("controller.Controller")
        self.__logger.info('creating an instance of Controller')
        self.__model = model
        self.__viewer = viewer
        self.__paused = False
        self.__next_tm = 0
        self.__date_from = make_date('1901/12/15') # TODO This seems to be the minimum date to be handled by date functions
        self.__date_to = make_date('2038/1/1')
        self.__location_filter = ""
        self.__where_clauses = {}
        self.__sort_clause = "exif_datetime ASC"
        self.publish_state = lambda x, y: None
        self.__keep_looping = True
        self.__location_filter = ''
        self.__tags_filter = ''
        self.__shutdown_complete = False

    @property
    def paused(self):
        """Get or set the current state for pausing image display. Setting paused to true
        will show the actual image as long paused is not set to false.
        """
        return self.__paused

    @paused.setter
    def paused(self, val:bool):
        self.__paused = val
        pic = self.__model.get_current_pics()[0] # only refresh left text
        self.__viewer.reset_name_tm(pic, val, side=0, pair=self.__model.get_current_pics()[1] is not None)

    def next(self):
        self.__next_tm = 0
        self.__viewer.reset_name_tm()

    def back(self):
        self.__model.set_next_file_to_previous_file()
        self.__next_tm = 0
        self.__viewer.reset_name_tm()

    def delete(self):
        self.__model.delete_file()
        self.back() # TODO check needed to avoid skipping one as record has been deleted from model.__file_list
        self.__next_tm = 0

    def set_show_text(self, txt_key=None, val="ON"):
        if val is True: # allow to be called with boolean from httpserver
            val = "ON"
        self.__viewer.set_show_text(txt_key, val)
        for (side, pic) in enumerate(self.__model.get_current_pics()):
            if pic is not None:
                self.__viewer.reset_name_tm(pic, self.paused, side, self.__model.get_current_pics()[1] is not None)

    def refresh_show_text(self):
        for (side, pic) in enumerate(self.__model.get_current_pics()):
            if pic is not None:
                self.__viewer.reset_name_tm(pic, self.paused, side, self.__model.get_current_pics()[1] is not None)

    @property
    def subdirectory(self):
        return self.__model.subdirectory

    @subdirectory.setter
    def subdirectory(self, dir):
        self.__model.subdirectory = dir
        self.__model.force_reload()
        self.__next_tm = 0

    @property
    def date_from(self):
        return self.__date_from

    @date_from.setter
    def date_from(self, val):
        try:
            self.__date_from = float(val)
        except ValueError:
            self.__date_from = make_date(val if len(val) > 0 else '1970/1/1')
        if len(val) > 0:
            self.__model.set_where_clause('date_from', "exif_datetime > {:.0f}".format(self.__date_from))
        else:
            self.__model.set_where_clause('date_from') # remove from where_clause
        self.__model.force_reload()
        self.__next_tm = 0

    @property
    def date_to(self):
        return self.__date_to

    @date_to.setter
    def date_to(self, val):
        try:
            self.__date_to = float(val)
        except ValueError:
            self.__date_to = make_date(val if len(val) > 0 else '2038/1/1')
        if len(val) > 0:
            self.__model.set_where_clause('date_to', "exif_datetime < {:.0f}".format(self.__date_to))
        else:
            self.__model.set_where_clause('date_to') # remove from where_clause
        self.__model.force_reload()
        self.__next_tm = 0

    @property
    def display_is_on(self):
        return self.__viewer.display_is_on

    @display_is_on.setter
    def display_is_on(self, on_off):
        self.__viewer.display_is_on = on_off

    @property
    def shuffle(self):
        return self.__model.shuffle

    @shuffle.setter
    def shuffle(self, val:bool):
        self.__model.shuffle = val
        self.__model.force_reload()
        self.__next_tm = 0

    @property
    def fade_time(self):
        return self.__model.fade_time

    @fade_time.setter
    def fade_time(self, time):
        self.__model.fade_time = float(time)
        self.__next_tm = 0

    @property
    def time_delay(self):
        return self.__model.time_delay

    @time_delay.setter
    def time_delay(self, time):
        time = float(time) # convert string before comparison
        # might break it if too quick
        if time < 5.0:
            time = 5.0
        self.__model.time_delay = time
        self.__next_tm = 0

    @property
    def brightness(self):
        return self.__viewer.get_brightness()

    @brightness.setter
    def brightness(self, val):
        self.__viewer.set_brightness(float(val))

    @property
    def location_filter(self):
        return self.__location_filter

    @location_filter.setter
    def location_filter(self, val):
        self.__location_filter = val
        if len(val) > 0:
            self.__model.set_where_clause('location_filter', "location LIKE '%{}%'".format(val))
        else:
            self.__model.set_where_clause('location_filter') # remove from where_clause
        self.__model.force_reload()
        self.__next_tm = 0

    @property
    def tags_filter(self):
        return self.__tags_filter
    
    @tags_filter.setter
    def tags_filter(self, val):
        self.__tags_filter = val
        if len(val) > 0:
            self.__model.set_where_clause('tags_filter', "tags LIKE '%{}%'".format(val))
        else:
            self.__model.set_where_clause('tags_filter') # remove from where_clause
        self.__model.force_reload()
        self.__next_tm = 0

    def text_is_on(self, txt_key):
        return self.__viewer.text_is_on(txt_key)

    def get_number_of_files(self):
        return self.__model.get_number_of_files()

    def get_directory_list(self):
        actual_dir, dir_list = self.__model.get_directory_list()
        return actual_dir, dir_list

    def get_current_path(self):
        (pic, _) = self.__model.get_current_pics()
        return pic.fname

    def loop(self): #TODO exit loop gracefully and call image_cache.stop()
        # catch ctrl-c
        signal.signal(signal.SIGINT, self.__signal_handler)

        #next_check_tm = time.time() + self.__model.get_model_config()['check_dir_tm']
        while self.__keep_looping:

            #if self.__next_tm == 0: #TODO double check why these were set when next_tm == 0
            #    time_delay = 1 # must not be 0
            #    fade_time = 1 # must not be 0
            #else:
            time_delay = self.__model.time_delay
            fade_time = self.__model.fade_time

            tm = time.time()
            pics = None #get_next_file returns a tuple of two in case paired portraits have been specified
            if not self.paused and tm > self.__next_tm:
                self.__next_tm = tm + self.__model.time_delay
                pics = self.__model.get_next_file()
                if pics[0] is None:
                    self.__next_tm = 0 # skip this image file moved or otherwise not on db
                    pics = None # signal slideshow_is_running not to load new image
                else:
                    image_attr = {}
                    for key in self.__model.get_model_config()['image_attr']:
                        if key == 'PICFRAME GPS':
                            image_attr['latitude'] = pics[0].latitude
                            image_attr['longitude'] = pics[0].longitude
                        elif key == 'PICFRAME LOCATION':
                            image_attr['location'] = pics[0].location
                        else:
                            field_name = self.__model.EXIF_TO_FIELD[key]
                            image_attr[key] = pics[0].__dict__[field_name] #TODO nicer using namedtuple for Pic
                    self.publish_state(pics[0].fname, image_attr)
            if self.__viewer.is_in_transition() == False: # safe to do long running tasks
                self.__model.pause_looping(True)
            else:
                self.__model.pause_looping(False) #TODO only need to set this once rather than every loop
            (loop_running, skip_image) = self.__viewer.slideshow_is_running(pics, time_delay, fade_time, self.__paused)
            if not loop_running:
                break
            if skip_image:
                self.__next_tm = 0
        self.__shutdown_complete = True

    def start(self):
        self.__viewer.slideshow_start()

    def stop(self):
        self.__keep_looping = False
        while not self.__shutdown_complete:
            time.sleep(0.05) # block until main loop has stopped
        self.__model.stop_image_chache() # close db tidily (blocks till closed)
        self.__viewer.slideshow_stop() # do this last
    
    def __signal_handler(self, sig, frame):
        print('You pressed Ctrl-c!')
        self.__shutdown_complete = True
        self.stop()



import pi3d
from pi3d.Texture import MAX_SIZE
import math
import time
import subprocess
import logging
import os
import numpy as np
from PIL import Image, ImageFilter, ImageFile
from picframe import mat_image

# supported display modes for display switch
dpms_mode = ("unsupported", "pi", "x_dpms")

# utility functions with no dependency on ViewerDisplay properties
def txt_to_bit(txt):
    txt_map = {"title":1, "caption":2, "name":4, "date":8, "location":16, "folder":32}
    if txt in txt_map:
        return txt_map[txt]
    return 0

def parse_show_text(txt):
    show_text = 0
    txt = txt.lower()
    for txt_key in ("title", "caption", "name", "date", "location", "folder"):
        if txt_key in txt:
            show_text |= txt_to_bit(txt_key)
    return show_text

class ViewerDisplay:

    def __init__(self, config):
        self.__logger = logging.getLogger("viewer_display.ViewerDisplay")
        self.__blur_amount = config['blur_amount']
        self.__blur_zoom = config['blur_zoom']
        self.__blur_edges = config['blur_edges']
        self.__edge_alpha = config['edge_alpha']

        self.__mat_images, self.__mat_images_tol = self.__get_mat_image_control_values(config['mat_images'])
        self.__mat_type = config['mat_type']
        self.__outer_mat_color = config['outer_mat_color']
        self.__inner_mat_color = config['inner_mat_color']
        self.__outer_mat_border = config['outer_mat_border']
        self.__inner_mat_border = config['inner_mat_border']
        self.__outer_mat_use_texture = config['outer_mat_use_texture']
        self.__inner_mat_use_texture = config['inner_mat_use_texture']
        self.__mat_resource_folder = os.path.expanduser(config['mat_resource_folder'])

        self.__fps = config['fps']
        self.__background = config['background']
        self.__blend_type = {"blend":0.0, "burn":1.0, "bump":2.0}[config['blend_type']]
        self.__font_file = os.path.expanduser(config['font_file'])
        self.__shader = os.path.expanduser(config['shader'])
        self.__show_text_tm = config['show_text_tm']
        self.__show_text_fm = config['show_text_fm']
        self.__show_text_sz = config['show_text_sz']
        self.__show_text = parse_show_text(config['show_text'])
        self.__text_justify = config['text_justify'].upper()
        self.__fit = config['fit']
        self.__auto_resize = config['auto_resize']
        self.__kenburns = config['kenburns']
        if self.__kenburns:
            self.__kb_up = True
            self.__fit = False
            self.__blur_edges = False
        if self.__blur_zoom < 1.0:
            self.__blur_zoom = 1.0
        self.__display_x = int(config['display_x'])
        self.__display_y = int(config['display_y'])
        self.__display_w = None if config['display_w'] is None else int(config['display_w'])
        self.__display_h = None if config['display_h'] is None else int(config['display_h'])
        self.__use_glx = config['use_glx']
        self.__codepoints = config['codepoints']
        self.__alpha = 0.0 # alpha - proportion front image to back
        self.__delta_alpha = 1.0
        self.__display = None
        self.__slide = None
        self.__xstep = None
        self.__ystep = None
        self.__text = None
        self.__textblocks = None
        self.__text_bkg = None
        self.__sfg = None # slide for background
        self.__sbg = None # slide for foreground
        self.__next_tm = 0.0
        self.__name_tm = 0.0
        self.__in_transition = False
        self.__matter = None
        ImageFile.LOAD_TRUNCATED_IMAGES = True # occasional damaged file hangs app

    @property
    def display_is_on(self):
        try: # vcgencmd only applies to raspberry pi
            state = str(subprocess.check_output(["vcgencmd", "display_power"]))
            if (state.find("display_power=1") != -1):
                return True
            else:
                return False
        except Exception as e:
            self.__logger.debug("Display ON/OFF is vcgencmd, but an error occurred")
            self.__logger.debug("Cause: %s", e)
            try: # try xset on linux, DPMS has to be enabled
                output = subprocess.check_output(["xset" , "-display", ":0", "-q"])
                if output.find(b'Monitor is On') != -1:
                    return True
                else:
                    return False
            except Exception as e:
                self.__logger.debug("Display ON/OFF is X with dpms enabled, but an error occurred")
                self.__logger.debug("Cause: %s", e)
                self.__logger.warning("Display ON/OFF is not supported for this platform.")
        return True

    @display_is_on.setter
    def display_is_on(self, on_off):
        try: # vcgencmd only applies to raspberry pi
            if on_off == True:
                subprocess.call(["vcgencmd", "display_power", "1"])
            else:
                subprocess.call(["vcgencmd", "display_power", "0"])
        except Exception as e:
            self.__logger.debug("Display ON/OFF is vcgencmd, but an error occured")
            self.__logger.debug("Cause: %s", e)
            try: # try xset on linux, DPMS has to be enabled
                if on_off == True:
                    subprocess.call(["xset" , "-display", ":0", "dpms", "force", "on"])
                else:
                    subprocess.call(["xset" , "-display", ":0", "dpms", "force", "off"])
            except Exception as e:
                self.__logger.debug("Display ON/OFF is xset via dpms, but an error occured")
                self.__logger.debug("Cause: %s", e)
                self.__logger.warning("Display ON/OFF is not supported for this platform.")

    def set_show_text(self, txt_key=None, val="ON"):
        if txt_key is None:
            self.__show_text = 0 # no arguments signals turning all off
        else:
            bit = txt_to_bit(txt_key) # convert field name to relevant bit 1,2,4,8,16 etc
            if val == "ON":
                self.__show_text |= bit # turn it on
            else: #TODO anything else ok to turn it off?
                bits = 65535 ^ bit
                self.__show_text &= bits # turn it off

    def text_is_on(self, txt_key):
        return self.__show_text & txt_to_bit(txt_key)

    def reset_name_tm(self, pic=None, paused=None, side=0, pair=False):
        # only extend i.e. if after initial fade in
        if pic is not None and paused is not None: # text needs to be refreshed
            self.__make_text(pic, paused, side, pair)
        self.__name_tm = max(self.__name_tm, time.time() + self.__show_text_tm)

    def set_brightness(self, val):
        self.__slide.unif[55] = val # take immediate effect

    def get_brightness(self):
        return float("{:.2f}".format(self.__slide.unif[55])) # TODO There seems to be a rounding issue. set 0.77 get 0.7699999809265137


    def __check_heif_then_open(self, fname):
        ext = os.path.splitext(fname)[1].lower()
        if ext in ('.heif','.heic'):
            try:
                import pyheif

                heif_file = pyheif.read(fname)
                image = Image.frombytes(heif_file.mode, heif_file.size, heif_file.data,
                                        "raw", heif_file.mode, heif_file.stride)
                if image.mode not in ("RGB", "RGBA"):
                    image = image.convert("RGB")
                return image
            except:
                self.__logger.warning("Failed attempt to convert %s \n** Have you installed pyheif? **", fname)
        else:
            try:
                image = Image.open(fname)
                if image.mode not in ("RGB", "RGBA"): # mat system needs RGB or more
                    image = image.convert("RGB")
            except: # for whatever reason
                image = None
            return image

    # Concatenate the specified images horizontally. Clip the taller
    # image to the height of the shorter image.
    def __create_image_pair(self, im1, im2):
        sep = 8 # separation between the images
        # scale widest image to same width as narrower to avoid drastic cropping on mismatched images
        if im1.width > im2.width:
            im1 = im1.resize((im2.width, int(im1.height * im2.width / im1.width)), resample=Image.BICUBIC)
        else:
            im2 = im2.resize((im1.width, int(im2.height * im1.width / im2.width)), resample=Image.BICUBIC)
        dst = Image.new('RGB', (im1.width + im2.width + sep, min(im1.height, im2.height)))
        dst.paste(im1, (0, 0))
        dst.paste(im2, (im1.width + sep, 0))
        return dst

    def __orientate_image(self, im, orientation):
        if orientation == 2:
            im = im.transpose(Image.FLIP_LEFT_RIGHT)
        elif orientation == 3:
            im = im.transpose(Image.ROTATE_180) # rotations are clockwise
        elif orientation == 4:
            im = im.transpose(Image.FLIP_TOP_BOTTOM)
        elif orientation == 5:
            im = im.transpose(Image.FLIP_LEFT_RIGHT).transpose(Image.ROTATE_90)
        elif orientation == 6:
            im = im.transpose(Image.ROTATE_270)
        elif orientation == 7:
            im = im.transpose(Image.FLIP_LEFT_RIGHT).transpose(Image.ROTATE_270)
        elif orientation == 8:
            im = im.transpose(Image.ROTATE_90)
        return im


    def __get_mat_image_control_values(self, mat_images_value):
        on = True
        val = 0.01
        org_val = str(mat_images_value).lower()
        if org_val in ('true', 'yes', 'on'):
            val = -1
        elif org_val in ('false', 'no', 'off'):
            on = False
        else:
            try:
                val = float(org_val)
            except:
                self.__logger.warning("Invalid value for config option 'mat_images'. Using default.")
        return(on, val)


    def __get_aspect_diff(self, screen_size, image_size):
        screen_aspect = screen_size[0] / screen_size[1]
        image_aspect = image_size[0] / image_size[1]

        if screen_aspect > image_aspect:
            diff_aspect = 1 - (image_aspect / screen_aspect)
        else:
            diff_aspect = 1 - (screen_aspect / image_aspect)

        return (screen_aspect, image_aspect, diff_aspect)


    def __tex_load(self, pics, size=None):
        try:
            if self.__mat_images and self.__matter == None:
                self.__matter = mat_image.MatImage(
                    display_size = (self.__display.width , self.__display.height),
                    resource_folder=self.__mat_resource_folder,
                    mat_type = self.__mat_type,
                    outer_mat_color = self.__outer_mat_color,
                    inner_mat_color = self.__inner_mat_color,
                    outer_mat_border = self.__outer_mat_border,
                    inner_mat_border = self.__inner_mat_border,
                    outer_mat_use_texture = self.__outer_mat_use_texture,
                    inner_mat_use_texture = self.__inner_mat_use_texture)

            # Load the image(s) and correct their orientation as necessary
            if pics[0]:
                im = self.__check_heif_then_open(pics[0].fname)
                if pics[0].orientation != 1:
                     im = self.__orientate_image(im, pics[0].orientation)
                if im is None:
                    return None
            if pics[1]:
                im2 = self.__check_heif_then_open(pics[1].fname)
                if pics[1].orientation != 1:
                     im2 = self.__orientate_image(im2, pics[1].orientation)

            screen_aspect, image_aspect, diff_aspect = self.__get_aspect_diff(size, im.size)

            if self.__mat_images and diff_aspect > self.__mat_images_tol:
                if not pics[1]:
                    im = self.__matter.mat_image((im,))
                else:
                    im = self.__matter.mat_image((im, im2))
            else:
                if pics[1]: #i.e portrait pair
                    im = self.__create_image_pair(im, im2)



            (w, h) = im.size
            max_dimension = MAX_SIZE # TODO changing MAX_SIZE causes serious crash on linux laptop!
            if not self.__auto_resize: # turned off for 4K display - will cause issues on RPi before v4
                max_dimension = 3840 # TODO check if mipmapping should be turned off with this setting.
            if w > max_dimension:
                im = im.resize((max_dimension, int(h * max_dimension / w)), resample=Image.BICUBIC)
            elif h > max_dimension:
                im = im.resize((int(w * max_dimension / h), max_dimension), resample=Image.BICUBIC)

            screen_aspect, image_aspect, diff_aspect = self.__get_aspect_diff(size, im.size)

            if self.__blur_edges and size:
                if diff_aspect > 0.01:
                    (sc_b, sc_f) = (size[1] / im.size[1], size[0] / im.size[0])
                    if screen_aspect > image_aspect:
                        (sc_b, sc_f) = (sc_f, sc_b) # swap round
                    (w, h) =  (round(size[0] / sc_b / self.__blur_zoom), round(size[1] / sc_b / self.__blur_zoom))
                    (x, y) = (round(0.5 * (im.size[0] - w)), round(0.5 * (im.size[1] - h)))
                    box = (x, y, x + w, y + h)
                    blr_sz = (int(x * 512 / size[0]) for x in size)
                    im_b = im.resize(size, resample=0, box=box).resize(blr_sz)
                    im_b = im_b.filter(ImageFilter.GaussianBlur(self.__blur_amount))
                    im_b = im_b.resize(size, resample=Image.BICUBIC)
                    im_b.putalpha(round(255 * self.__edge_alpha))  # to apply the same EDGE_ALPHA as the no blur method.
                    im = im.resize((int(x * sc_f) for x in im.size), resample=Image.BICUBIC)
                    """resize can use Image.LANCZOS (alias for Image.ANTIALIAS) for resampling
                    for better rendering of high-contranst diagonal lines. NB downscaled large
                    images are rescaled near the start of this try block if w or h > max_dimension
                    so those lines might need changing too.
                    """
                    im_b.paste(im, box=(round(0.5 * (im_b.size[0] - im.size[0])),
                                        round(0.5 * (im_b.size[1] - im.size[1]))))
                    im = im_b # have to do this as paste applies in place
            tex = pi3d.Texture(im, blend=True, m_repeat=True, automatic_resize=self.__auto_resize,
                                free_after_load=True)
            #tex = pi3d.Texture(im, blend=True, m_repeat=True, automatic_resize=config.AUTO_RESIZE,
            #                    mipmap=config.AUTO_RESIZE, free_after_load=True) # poss try this if still some artifacts with full resolution
        except Exception as e:
            self.__logger.warning("Can't create tex from file: \"%s\" or \"%s\"", pics[0].fname, pics[1])
            self.__logger.warning("Cause: %s", e)
            tex = None
            #raise # only re-raise errors here while debugging
        return tex

    def __sanitize_string(self, path_name):
        name = os.path.basename(path_name)
        name = ''.join([c for c in name if c in self.__codepoints])
        return name

    def __make_text(self, pic, paused, side=0, pair=False):
        # if side 0 and pair False then this is a full width text and put into
        # __textblocks[0] otherwise it is half width and put into __textblocks[position]
        info_strings = []
        if pic is not None and (self.__show_text > 0 or paused): #was SHOW_TEXT_TM > 0.0
            if (self.__show_text & 1) == 1 and pic.title is not None: # title
                info_strings.append(self.__sanitize_string(pic.title))
            if (self.__show_text & 2) == 2 and pic.caption is not None: # caption
                info_strings.append(self.__sanitize_string(pic.caption))
            if (self.__show_text & 4) == 4: # name
                info_strings.append(self.__sanitize_string(pic.fname))
            if (self.__show_text & 8) == 8 and pic.exif_datetime > 0: # date
                fdt = time.strftime(self.__show_text_fm, time.localtime(pic.exif_datetime))
                info_strings.append(fdt)
            if (self.__show_text & 16) == 16 and pic.location is not None: # location
                info_strings.append(pic.location) #TODO need to sanitize and check longer than 0 for real
            if (self.__show_text & 32) == 32: # folder
                info_strings.append(self.__sanitize_string(os.path.basename(os.path.dirname(pic.fname))))
            if paused:
                info_strings.append("PAUSED")
        final_string = " • ".join(info_strings)
        if len(final_string) > 0:
            final_string += " " # TODO fix in pi3d.TextBlock if last line too long the space will trigger wrapping

        block = self.__textblocks[side] # alias for brevity below
        if side == 0 and not pair:
            c_rng = self.__display.width - 100 # range for x loc from L to R justified
            x = int(c_rng * (block.justify - 0.5))
        else:
            c_rng = self.__display.width * 0.5 - 100 # range for x loc from L to R justified
            x = int(c_rng * (block.justify - 1) - 50) if side == 0 else int(c_rng * block.justify + 50)
        block.set_text(text_format=final_string, wrap_pixels=c_rng)
        #TODO next few lines are a way to find which char will be drawn and the position of lowest
        # char. This functionality could be placed in pi3d.TextBlock
        block_uvs = block._text_manager.uv[block._buffer_index:(block._buffer_index + block.char_count),:]
        ix = np.where((block_uvs[:,0] != 0.0) & (block_uvs[:,1] != 0.0))[0] # block_uvs is view into larger array held by text
        if len(ix) > 0: # i.e. something will be drawn. ix is array of indices into char_offsets array to draw
            adj_y = block.y + block.char_offsets[ix,1].min() + self.__display.height // 2
            block.set_position(x=x, y=(block.y - adj_y + self.__show_text_sz))

    def is_in_transition(self):
        return self.__in_transition

    def slideshow_start(self):
        self.__display = pi3d.Display.create(x=self.__display_x, y=self.__display_y,
              w=self.__display_w, h=self.__display_h, frames_per_second=self.__fps,
              display_config=pi3d.DISPLAY_CONFIG_HIDE_CURSOR, background=self.__background, use_glx=self.__use_glx)
        camera = pi3d.Camera(is_3d=False)
        shader = pi3d.Shader(self.__shader)
        self.__slide = pi3d.Sprite(camera=camera, w=self.__display.width, h=self.__display.height, z=5.0)
        self.__slide.set_shader(shader)
        self.__slide.unif[47] = self.__edge_alpha
        self.__slide.unif[54] = float(self.__blend_type)
        self.__slide.unif[55] = 1.0 #brightness
        # PointText and TextBlock. If SHOW_NAMES_TM <= 0 then this is just used for no images message
        grid_size = math.ceil(len(self.__codepoints) ** 0.5)
        font = pi3d.Font(self.__font_file, codepoints=self.__codepoints, grid_size=grid_size, shadow_radius=4.0,
                        shadow=(0,0,0,128))
        self.__text = pi3d.PointText(font, camera, max_chars=400, point_size=self.__show_text_sz)
        self.__textblocks = []
        if self.__text_justify == 'C':
            justify = 0.5
        elif self.__text_justify == 'R':
            justify = 1.0
        else:
            justify = 0.0
        for i in range(2):
            # x position will now be recalculated for each make_text call
            block = pi3d.TextBlock(x=0, y=-int(self.__display.height * 0.4),
                                z=0.1, rot=0.0, char_count=199,
                                text_format="{}".format(" "), size=0.99, spacing="F",
                                space=0.02, colour=(1.0, 1.0, 1.0, 1.0), justify=justify)
            self.__text.add_text_block(block)
            self.__textblocks.append(block)
        bkg_ht = self.__display.height // 4
        text_bkg_array = np.zeros((bkg_ht, 1, 4), dtype=np.uint8)
        text_bkg_array[:,:,3] = np.linspace(0, 120, bkg_ht).reshape(-1, 1)
        text_bkg_tex = pi3d.Texture(text_bkg_array, blend=True, mipmap=False, free_after_load=True)

        back_shader = pi3d.Shader("uv_flat")
        self.__text_bkg = pi3d.Sprite(w=self.__display.width, h=bkg_ht, y=-int(self.__display.height) // 2 + bkg_ht // 2, z=4.0)
        self.__text_bkg.set_draw_details(back_shader, [text_bkg_tex])


    def slideshow_is_running(self, pics=None, time_delay = 200.0, fade_time = 10.0, paused=False):
        tm = time.time()
        if pics is not None:
            self.__sbg = self.__sfg # if the first tex_load fails then __sfg might be Null TODO should fn return if None?
            self.__next_tm = tm + time_delay
            self.__name_tm = tm + fade_time + float(self.__show_text_tm) # text starts after slide transition
            new_sfg = self.__tex_load(pics, (self.__display.width, self.__display.height))
            if new_sfg is not None: # this is a possible return value which needs to be caught
                self.__sfg = new_sfg
            else:
                return (True, False) # return early
            self.__alpha = 0.0
            self.__delta_alpha = 1.0 / (self.__fps * fade_time) # delta alpha
            # set the file name as the description
            if self.__show_text_tm > 0.0:
                for i, pic in enumerate(pics):
                    self.__make_text(pic, paused, i, pics[1] is not None) # send even if pic is None to clear previous text
                self.__text.regen()
            else: # could have a NO IMAGES selected and being drawn
                for block in self.__textblocks:
                    block.set_text(text_format="{}".format(" "))
                    block.colouring.set_colour(alpha=0.0)
                self.__text.regen()

            if self.__sbg is None: # first time through
                self.__sbg = self.__sfg
            self.__slide.set_textures([self.__sfg, self.__sbg])
            self.__slide.unif[45:47] = self.__slide.unif[42:44] # transfer front width and height factors to back
            self.__slide.unif[51:53] = self.__slide.unif[48:50] # transfer front width and height offsets
            wh_rat = (self.__display.width * self.__sfg.iy) / (self.__display.height * self.__sfg.ix)
            if (wh_rat > 1.0 and self.__fit) or (wh_rat <= 1.0 and not self.__fit):
                sz1, sz2, os1, os2 = 42, 43, 48, 49
            else:
                sz1, sz2, os1, os2 = 43, 42, 49, 48
                wh_rat = 1.0 / wh_rat
            self.__slide.unif[sz1] = wh_rat
            self.__slide.unif[sz2] = 1.0
            self.__slide.unif[os1] = (wh_rat - 1.0) * 0.5
            self.__slide.unif[os2] = 0.0
            if self.__kenburns:
                self.__xstep, self.__ystep = (self.__slide.unif[i] * 2.0 / (time_delay - fade_time) for i in (48, 49))
                self.__slide.unif[48] = 0.0
                self.__slide.unif[49] = 0.0
                #self.__kb_up = not self.__kb_up # just go in one direction

        if self.__kenburns and self.__alpha >= 1.0:
            t_factor = time_delay - fade_time - self.__next_tm + tm
            #t_factor = self.__next_tm - tm
            #if self.__kb_up:
            #    t_factor = time_delay - t_factor
            # add exponentially smoothed tweening in case of timing delays etc. to avoid 'jumps'
            self.__slide.unif[48] = self.__slide.unif[48] * 0.95 + self.__xstep * t_factor * 0.05
            self.__slide.unif[49] = self.__slide.unif[49] * 0.95 + self.__ystep * t_factor * 0.05

        if self.__alpha < 1.0: # transition is happening
            self.__alpha += self.__delta_alpha
            if self.__alpha > 1.0:
                self.__alpha = 1.0
            self.__slide.unif[44] = self.__alpha * self.__alpha * (3.0 - 2.0 * self.__alpha)

        if (self.__next_tm - tm) < 5.0 or self.__alpha < 1.0:
            self.__in_transition = True # set __in_transition True a few seconds *before* end of previous slide
        else: # no transition effect safe to update database, resuffle etc
            self.__in_transition = False

        self.__slide.draw()

        if self.__alpha >= 1.0 and tm < self.__name_tm:
            # this sets alpha for the TextBlock from 0 to 1 then back to 0
            dt = (self.__show_text_tm - self.__name_tm + tm + 0.1) / self.__show_text_tm
            ramp_pt = max(4.0, self.__show_text_tm / 4.0)
            alpha = max(0.0, min(1.0, ramp_pt * (self.__alpha- abs(1.0 - 2.0 * dt)))) # cap text alpha at image alpha
            for block in self.__textblocks:
                block.colouring.set_colour(alpha=alpha)
            self.__text.regen()
            self.__text_bkg.set_alpha(alpha)
            txt_len = 0
            for block in self.__textblocks:
                txt_len += len(block.text_format.strip())
            if txt_len > 0: #only draw background if text there
                self.__text_bkg.draw()


        self.__text.draw()
        return (self.__display.loop_running(), False) # now returns tuple with skip image flag added

    def slideshow_stop(self):
        self.__display.destroy()


import pytest
import logging

import os


from picframe import model

logger = logging.getLogger("test_model")
logger.setLevel(logging.DEBUG)

def test_model_init():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    mqtt = m.get_mqtt_config()
    assert mqtt['server'] == 'home'
    viewer = m.get_viewer_config()
    assert viewer['test_key'] == 'test_value'
  

def test_for_file_changes():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    m.subdirectory = 'testdir'
    testfile = os.path.expanduser(m.get_model_config()['pic_dir']) + "/"+ 'testdir' + "/testfile.jpg"
    assert m.check_for_file_changes() == False
    os.mknod(testfile)
    assert m.check_for_file_changes() == True
    os.remove(testfile)

def test_get_files():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    num = m.get_number_of_files()
    assert num == 443 

def test_get_next_file():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    file1 = m.get_next_file()
    file2 = m.get_next_file()
    assert file1 != file2

def test_get_next_file_whole_loop():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    num = m.get_number_of_files()
    m.shuffle = False
    file1 = m.get_next_file()
    file2 = None
    for _ in range(0, num):
        file2 = m.get_next_file()
    assert file1 == file2

def test_get_next_file_no_file_in_range():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    file, orientation, image_attr = m.get_next_file((1990,1,1), (1990,1,2))
    assert file == '/home/pi/.local/picframe/data/PictureFrame2020img.jpg'
    assert orientation == 1
    assert image_attr == {
        'latitude':None,
        'longitude':None,
        'EXIF FNumber':None,
        'EXIF ExposureTime':None,
        'EXIF ISOSpeedRatings':None,
        'EXIF FocalLength':None,
        'EXIF DateTimeOriginal':None,
        'Image Model':None}

def test_get_file_for_empty_dir():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    m.subdirectory = 'testdir'
    file, orientation, image_attr  = m.get_next_file()
    assert file == '/home/pi/.local/picframe/data/PictureFrame2020img.jpg'
    assert orientation == 1
    assert image_attr == {
        'latitude':None,
        'longitude':None,
        'EXIF FNumber':None,
        'EXIF ExposureTime':None,
        'EXIF ISOSpeedRatings':None,
        'EXIF FocalLength':None,
        'EXIF DateTimeOriginal':None,
        'Image Model':None}

def test_getter_setter_fade_time():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    assert m.fade_time == 10.0
    m.fade_time = 20.0
    assert m.fade_time == 20.0

def test_getter_setter_time_delay():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    assert m.time_delay == 200.0
    m.time_delay = 21.0
    assert m.time_delay == 21.0

def test_getter_setter_subdirectory():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    assert m.subdirectory == ''
    m.subdirectory = 'testdir'
    assert m.subdirectory == 'testdir'

def test_getter_setter_shuffle():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    assert m.shuffle == True
    m.shuffle = False
    assert m.shuffle == False


def test_get_subdirectory_list():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    act_dir, dir_list = m.get_directory_list()
    assert act_dir == 'Pictures'
    assert dir_list[0] == 'Pictures'
    assert dir_list[1] == 'testdir'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
import os
import logging
import json
import threading
from functools import partial

try:
    from http.server import BaseHTTPRequestHandler, HTTPServer #py3
    import urllib.parse as urlparse
except ImportError:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer #py2
    import urlparse

EXTENSIONS = [".jpg", ".jpeg", ".png", ".heif", ".heic"]

def heif_to_jpg(fname):
    try:
        import pyheif
        from PIL import Image

        heif_file = pyheif.read(fname)
        image = Image.frombytes(heif_file.mode, heif_file.size, heif_file.data,
                                "raw", heif_file.mode, heif_file.stride)
        if image.mode not in ("RGB", "RGBA"):
            image = image.convert("RGB")
        image.save("/dev/shm/temp.jpg") # default 75% quality
        return "/dev/shm/temp.jpg"
    except:
        self.__logger.warning("Failed attempt to convert %s \n** Have you installed pyheif? **", fname)
        return "" # this will not render as a page and will generate error TODO serve specific page with explicit error

class RequestHandler(BaseHTTPRequestHandler):

    def do_GET(self):
        try:
            path_split = self.path.split("?")
            page_ok = False
            if len(path_split) == 1: # i.e. no ? - just serve index.html or image
                if path_split[0] != "/": # serve static page from html_path...
                    html_page = path_split[0].strip("/")
                else:
                    html_page = "index.html"
                _, extension = os.path.splitext(html_page)
                if html_page == "current_image" or extension.lower() in EXTENSIONS:
                    # NB homeassistant needs to pass url ending in an image extension
                    # in order to trigger streaming whatever is the currently showing image
                    content_type = "image"
                    page = self.server._controller.get_current_path()
                    _, extension = os.path.splitext(page) # as current_image may be heic
                    if extension.lower() in ('.heic', '.heif'):
                        page = heif_to_jpg(page)
                else:
                    page = os.path.join(self.server._html_path, html_page)
                    content_type = "text/html"
                page = urlparse.unquote(page)
                if os.path.isfile(page):
                    self.send_response(200)
                    self.send_header('Content-type', content_type)
                    #TODO check if html or js - in which case application/javascript
                    # really should filter out attempts to render all other file types (jpg etc?)
                    self.end_headers()
                    with open(page, "rb") as f:
                        page_bytes = f.read()
                        self.wfile.write(page_bytes)
                    self.connection.close()
                    page_ok = True
            else: # server type request - get or set info
                start_time = time.time()
                message = {}
                self.send_response(200)
                self.server._logger.debug('http request from: ' + self.client_address[0])

                for key, value in dict(urlparse.parse_qsl(path_split[1], True)).items():
                    self.send_header('Content-type', 'text')
                    self.end_headers()
                    if key == "all":
                        for subkey in self.server._setters:
                            message[subkey] = getattr(self.server._controller, subkey)
                    elif key in dir(self.server._controller):
                        if key in self.server._setters: # can info back from controller
                            message[key] = getattr(self.server._controller, key)
                        lwr_val = value.lower()
                        if lwr_val in ("true", "on", "yes"): # this only works for simple values *not* json style kwargs
                            value = True
                        elif lwr_val in ("false", "off", "no"):
                            value = False
                        try:
                            if key in self.server._setters:
                                setattr(self.server._controller, key, value)
                            else:
                                value = value.replace("\'", "\"") # only " permitted in json
                                # value must be json kwargs
                                getattr(self.server._controller, key)(**json.loads(value))
                        except Exception as e:
                            message['ERROR'] = 'Excepton:{}>{};'.format(key, e)

                    self.wfile.write(bytes(json.dumps(message), "utf8"))
                    self.connection.close()
                    page_ok = True

                self.server._logger.info(message)
                self.server._logger.debug("request finished in:  %s seconds" %
                              (time.time() - start_time))
            if not page_ok:
                self.send_response(404)
                self.connection.close()
        except Exception as e:
            self.server._logger.warning(e)
            self.send_response(400)
            self.connection.close()

        return


    def log_request(self, code):
        pass


    def do_POST(self):
        self.do_GET()


    def end_headers(self):
        try:
            super().end_headers()
        except BrokenPipeError as e:
            self.connection.close()
            self.server._logger.error('httpserver error: {}'.format(e))


class InterfaceHttp(HTTPServer):
    def __init__(self, controller, html_path, pic_dir, no_files_img, port=9000):
        super(InterfaceHttp, self).__init__(("0.0.0.0", port), RequestHandler)
        # NB name mangling throws a spanner in the works here!!!!!
        # *no* __dunders
        self._logger = logging.getLogger("simple_server.InterfaceHttp")
        self._logger.info("creating an instance of InterfaceHttp")
        self._controller = controller
        self._pic_dir = os.path.expanduser(pic_dir)
        self._no_files_img = os.path.expanduser(no_files_img)
        self._html_path = os.path.expanduser(html_path)
        self._setters = ["paused", "subdirectory", "date_from", "date_to",
                         "display_is_on", "shuffle", "fade_time", "time_delay",
                         "brightness", "location_filter"] #TODO can this be done with dir() and getattr() to avoid hard coding?
        self.__keep_looping = True
        self.__shutdown_completed = False
        t = threading.Thread(target=self.__loop)
        t.start()

    def __loop(self):
        while self.__keep_looping:
            self.handle_request()
            time.sleep(0.1)
        self.__shutdown_completed = True

    def stop(self):
        self.__keep_looping = False
        while not self.__shutdown_completed:
            time.sleep(0.05) # function blocking until loop stopped



from setuptools import setup, find_packages
from glob import glob
import versioneer

# read the contents of your README file
import os
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='picframe',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Picture frame viewer powered by raspberry with homeassistant integration',
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Multimedia :: Graphics :: Viewers',
      ],
      keywords='picframe viewer raspberry raspi homeassistant hass',
      url='https://github.com/helgeerbe/picframe',
      author='Paddy Gaunt, Jeff Godfrey, Helge Erbe',
      author_email='helge@erbehome.de',
      license='MIT',
      packages=find_packages(exclude=("test")),
      package_data={ '': [
        'data/*', 'data/**/*', 
        'config/*', 'config/**/*', 
        'html/*', 'html/**/*']},
      install_requires=[
        'Pillow',
        'ExifRead',
        'pi3d>=2.43',
        'PyYAML',
        'paho-mqtt',
        'IPTCInfo3',
        'numpy',
        'ninepatch'
      ],
      entry_points = {
        'console_scripts': ['picframe=picframe.start:main']
      },
      include_package_data=True,
      zip_safe=False)



import sqlite3
import os
import time
import logging
import threading
from picframe import get_image_meta

class ImageCache:

    EXTENSIONS = ['.png','.jpg','.jpeg','.heif','.heic']
    EXIF_TO_FIELD = {'EXIF FNumber': 'f_number',
                     'EXIF Make': 'make',
                     'Image Model': 'model',
                     'EXIF ExposureTime': 'exposure_time',
                     'EXIF ISOSpeedRatings': 'iso',
                     'EXIF FocalLength': 'focal_length',
                     'EXIF Rating': 'rating',
                     'EXIF LensModel': 'lens',
                     'EXIF DateTimeOriginal': 'exif_datetime',
                     'IPTC Keywords': 'tags',
                     'IPTC Caption/Abstract': 'caption',
                     'IPTC Object Name': 'title'}


    def __init__(self, picture_dir, db_file, geo_reverse, required_db_schema_version=1, portrait_pairs=False):
        self.__modified_folders = []
        self.__modified_files = []
        self.__logger = logging.getLogger("image_cache.ImageCache")
        self.__logger.debug('Creating an instance of ImageCache')
        self.__picture_dir = picture_dir
        self.__db_file = db_file
        self.__geo_reverse = geo_reverse
        self.__portrait_pairs = portrait_pairs #TODO have a function to turn this on and off?
        self.__db = self.__create_open_db(self.__db_file)
        self.__update_schema(required_db_schema_version)

        self.__keep_looping = True
        self.__pause_looping = False
        self.__shutdown_completed = False

        t = threading.Thread(target=self.__loop)
        t.start()


    def __loop(self):
        while self.__keep_looping:
            if not self.__pause_looping:
                self.update_cache()
                time.sleep(2.0)
            time.sleep(0.01)
        self.__db.commit() # close after update_cache finished for last time
        self.__db.close()
        self.__shutdown_completed = True


    def pause_looping(self, value):
        self.__pause_looping = value


    def stop(self):
        self.__keep_looping = False
        while not self.__shutdown_completed:
            time.sleep(0.05) # make function blocking to ensure staged shutdown


    def update_cache(self):
        """Update the cache database with new and/or modified files
        """

        self.__logger.debug('Updating cache')

        # If the current collection of updated files is empty, check for disk-based changes
        if not self.__modified_files:
            self.__logger.debug('No unprocessed files in memory, checking disk')
            self.__modified_folders = self.__get_modified_folders()
            self.__modified_files = self.__get_modified_files(self.__modified_folders)
            self.__logger.debug('Found %d new files on disk', len(self.__modified_files))

        # While we have files to process and looping isn't paused
        while self.__modified_files and not self.__pause_looping:
            file = self.__modified_files.pop(0)
            self.__logger.debug('Inserting: %s', file)
            self.__insert_file(file)

        # If we've process all files in the current collection, update the cached folder mod times
        if not self.__modified_files:
            self.__update_folder_modtimes(self.__modified_folders)
            self.__modified_folders.clear()

        # If looping is still not paused, remove any files or folders from the db that are no longer on disk
        if not self.__pause_looping:
            self.__purge_missing_files_and_folders()

        # Commit the current set of changes
        self.__db.commit()


    def query_cache(self, where_clause, sort_clause = 'exif_datetime ASC'):
        cursor = self.__db.cursor()
        cursor.row_factory = None # we don't want the "sqlite3.Row" setting from the db here...

        if not self.__portrait_pairs: # TODO SQL insertion? Does it matter in this app?
            sql = """SELECT file_id FROM all_data WHERE {0} ORDER BY {1}
                """.format(where_clause, sort_clause)
            return cursor.execute(sql).fetchall()
        else: # make two SELECTS
            sql = """SELECT
                        CASE
                            WHEN is_portrait = 0 THEN file_id
                            ELSE -1
                        END
                        FROM all_data WHERE {0} ORDER BY {1}
                                    """.format(where_clause, sort_clause)
            full_list = cursor.execute(sql).fetchall()
            sql = """SELECT file_id FROM all_data
                        WHERE ({0}) AND is_portrait = 1 ORDER BY {1}
                                    """.format(where_clause, sort_clause)
            pair_list = cursor.execute(sql).fetchall()
            newlist = []
            for i in range(len(full_list)):
                if full_list[i][0] != -1:
                    newlist.append(full_list[i])
                elif pair_list: #OK @rec - this is tidier and qicker!
                    elem = pair_list.pop(0)
                    if pair_list:
                        elem += pair_list.pop(0)
                    newlist.append(elem)
            return newlist


    def get_file_info(self, file_id):
        sql = "SELECT * FROM all_data where file_id = {0}".format(file_id)
        row = self.__db.execute(sql).fetchone()
        if row is not None and row['latitude'] is not None and row['longitude'] is not None and row['location'] is None:
            if self.__get_geo_location(row['latitude'], row['longitude']):
                row = self.__db.execute(sql).fetchone() # description inserted in table
        return row # NB if select fails (i.e. moved file) will return None

    def __get_geo_location(self, lat, lon): # TODO periodically check all lat/lon in meta with no location and try again
        location = self.__geo_reverse.get_address(lat, lon)
        if len(location) == 0:
            return False #TODO this will continue to try even if there is some permanant cause
        else:
            sql = "INSERT OR REPLACE INTO location (latitude, longitude, description) VALUES (?, ?, ?)"
            self.__db.execute(sql, (lat, lon, location))
            return True


    def __create_open_db(self, db_file):
        sql_folder_table = """
            CREATE TABLE IF NOT EXISTS folder (
                folder_id INTEGER NOT NULL PRIMARY KEY,
                name TEXT UNIQUE NOT NULL,
                last_modified REAL DEFAULT 0 NOT NULL
            )"""

        sql_file_table = """
            CREATE TABLE IF NOT EXISTS file (
                file_id INTEGER NOT NULL PRIMARY KEY,
                folder_id INTEGER NOT NULL,
                basename  TEXT NOT NULL,
                extension TEXT NOT NULL,
                last_modified REAL DEFAULT 0 NOT NULL,
                UNIQUE(folder_id, basename, extension)
            )"""

        sql_meta_table = """
            CREATE TABLE IF NOT EXISTS meta (
                file_id INTEGER NOT NULL PRIMARY KEY,
                orientation INTEGER DEFAULT 1 NOT NULL,
                exif_datetime REAL DEFAULT 0 NOT NULL,
                f_number REAL DEFAULT 0 NOT NULL,
                exposure_time TEXT,
                iso REAL DEFAULT 0 NOT NULL,
                focal_length TEXT,
                make TEXT,
                model TEXT,
                lens TEXT,
                rating INTEGER,
                latitude REAL,
                longitude REAL,
                width INTEGER DEFAULT 0 NOT NULL,
                height INTEGER DEFAULT 0 NOT NULL,
                title TEXT,
                caption TEXT,
                tags TEXT
            )"""

        sql_meta_index = """
            CREATE INDEX IF NOT EXISTS exif_datetime ON meta (exif_datetime)"""

        sql_location_table = """
            CREATE TABLE IF NOT EXISTS location (
                id INTEGER NOT NULL PRIMARY KEY,
                latitude REAL,
                longitude REAL,
                description TEXT,
                UNIQUE (latitude, longitude)
            )"""

        sql_db_info_table = """
            CREATE TABLE IF NOT EXISTS db_info (
                schema_version INTEGER NOT NULL
            )"""

        # Combine all important data in a single view for easy accesss
        # Although we can't control the layout of the view when using 'meta.*', we want it
        # all and that seems better than enumerating (and maintaining) each column here.
        sql_all_data_view = """
            CREATE VIEW IF NOT EXISTS all_data
            AS
            SELECT
                folder.name || "/" || file.basename || "." || file.extension AS fname,
                file.last_modified,
                meta.*,
                meta.height > meta.width as is_portrait,
                location.description as location
            FROM file
                INNER JOIN folder
                    ON folder.folder_id = file.folder_id
                LEFT JOIN meta
                    ON file.file_id = meta.file_id
                LEFT JOIN location
                    ON location.latitude = meta.latitude AND location.longitude = meta.longitude
            """

        # trigger to automatically delete file records when associated folder records are deleted
        sql_clean_file_trigger = """
            CREATE TRIGGER IF NOT EXISTS Clean_File_Trigger
            AFTER DELETE ON folder
            FOR EACH ROW
            BEGIN
                DELETE FROM file WHERE folder_id = OLD.folder_id;
            END"""

        # trigger to automatically delete meta records when associated file records are deleted
        sql_clean_meta_trigger = """
            CREATE TRIGGER IF NOT EXISTS Clean_Meta_Trigger
            AFTER DELETE ON file
            FOR EACH ROW
            BEGIN
                DELETE FROM meta WHERE file_id = OLD.file_id;
            END"""

        db = sqlite3.connect(db_file, check_same_thread=False) # writing only done in loop thread, reading in this so should be safe
        db.row_factory = sqlite3.Row # make results accessible by field name
        for item in (sql_folder_table, sql_file_table, sql_meta_table, sql_location_table, sql_meta_index,
                    sql_all_data_view, sql_db_info_table, sql_clean_file_trigger, sql_clean_meta_trigger):
            db.execute(item)

        return db


    def __update_schema(self, required_db_schema_version):
        sql_select = "SELECT schema_version from db_info"
        schema_version = self.__db.execute(sql_select).fetchone()
        schema_version = 1 if not schema_version else schema_version[0]

        # DB is newer than the application. The User needs to upgrade...
        if schema_version > required_db_schema_version:
            raise ValueError("Database schema is newer than the application. Update the application.")

        # Here, we need to update the db schema as necessary
        if schema_version < required_db_schema_version:
            pass

        # Finally, update the schema version stamp
        self.__db.execute('DELETE FROM db_info')
        self.__db.execute('INSERT INTO db_info VALUES(?)', (required_db_schema_version,))


    def __get_modified_folders(self):
        out_of_date_folders = []
        sql_select = "SELECT * FROM folder WHERE name = ?"
        for dir in [d[0] for d in os.walk(self.__picture_dir)]:
            mod_tm = int(os.stat(dir).st_mtime)
            found = self.__db.execute(sql_select, (dir,)).fetchone()
            if not found or found['last_modified'] < mod_tm:
                out_of_date_folders.append((dir, mod_tm))
        return out_of_date_folders


    def __get_modified_files(self, modified_folders):
        out_of_date_files = []
        sql_select = "SELECT fname, last_modified FROM all_data WHERE fname = ? and last_modified >= ?"
        for dir,date in modified_folders:
            for file in os.listdir(dir):
                base, extension = os.path.splitext(file)
                if (extension.lower() in ImageCache.EXTENSIONS
                        and not '.AppleDouble' in dir and not file.startswith('.')): # have to filter out all the Apple junk
                    full_file = os.path.join(dir, file)
                    mod_tm =  os.path.getmtime(full_file)
                    found = self.__db.execute(sql_select, (full_file,mod_tm)).fetchone()
                    if not found:
                        out_of_date_files.append(full_file)
        return out_of_date_files


    def __insert_file(self, file):
        file_insert = "INSERT OR REPLACE INTO file(folder_id, basename, extension, last_modified) VALUES((SELECT folder_id from folder where name = ?), ?, ?, ?)"
        folder_insert = "INSERT OR IGNORE INTO folder(name) VALUES(?)"
        mod_tm =  os.path.getmtime(file)
        dir, file_only = os.path.split(file)
        base, extension = os.path.splitext(file_only)

        # Get the file's meta info and build the INSERT statement dynamically
        meta = self.__get_exif_info(file)
        meta_insert = self.__get_meta_sql_from_dict(meta)
        vals = list(meta.values())
        vals.insert(0, file)

        # Insert this file's info into the folder, file, and meta tables
        self.__db.execute(folder_insert, (dir,))
        self.__db.execute(file_insert, (dir, base, extension.lstrip("."), mod_tm))
        self.__db.execute(meta_insert, vals)


    def __update_folder_modtimes(self, folder_collection):
        update_data = []
        sql = "UPDATE folder SET last_modified = ? WHERE name = ?"
        for folder, modtime in folder_collection:
            update_data.append((modtime, folder))
        self.__db.executemany(sql, update_data)


    def __get_meta_sql_from_dict(self, dict):
        columns = ', '.join(dict.keys())
        ques = ', '.join('?' * len(dict.keys()))
        return 'INSERT OR REPLACE INTO meta(file_id, {0}) VALUES((SELECT file_id from all_data where fname = ?), {1})'.format(columns, ques)


    def __purge_missing_files_and_folders(self):
        # Find folders in the db that are no longer on disk
        folder_id_list = []
        for row in self.__db.execute('SELECT folder_id, name from folder'):
            if not os.path.exists(row['name']):
                folder_id_list.append([row['folder_id']])

        # Delete any non-existent folders from the db. Note, this will automatically
        # remove orphaned records from the 'file' and 'meta' tables
        if len(folder_id_list):
            self.__db.executemany('DELETE FROM folder WHERE folder_id = ?', folder_id_list)

        # Find files in the db that are no longer on disk
        file_id_list = []
        for row in self.__db.execute('SELECT file_id, fname from all_data'):
            if not os.path.exists(row['fname']):
                file_id_list.append([row['file_id']])

        # Delete any non-existent files from the db. Note, this will automatically
        # remove matching records from the 'meta' table as well.
        if len(file_id_list):
            self.__db.executemany('DELETE FROM file WHERE file_id = ?', file_id_list)


    def __get_exif_info(self, file_path_name):
        exifs = get_image_meta.GetImageMeta(file_path_name)
        # Dict to store interesting EXIF data
        # Note, the 'key' must match a field in the 'meta' table
        e = {}

        e['orientation'] = exifs.get_orientation()

        width, height = exifs.get_size()
        if e['orientation'] in (5, 6, 7, 8):
            width, height = height, width # swap values
        e['width'] = width
        e['height'] = height


        e['f_number'] = exifs.get_exif('EXIF FNumber')
        e['make'] = exifs.get_exif('EXIF Make')
        e['model'] = exifs.get_exif('Image Model')
        e['exposure_time'] = exifs.get_exif('EXIF ExposureTime')
        e['iso'] =  exifs.get_exif('EXIF ISOSpeedRatings')
        e['focal_length'] =  exifs.get_exif('EXIF FocalLength')
        e['rating'] = exifs.get_exif('EXIF Rating')
        e['lens'] = exifs.get_exif('EXIF LensModel')
        val = exifs.get_exif('EXIF DateTimeOriginal')
        if val != None:
            # Remove any subsecond portion of the DateTimeOriginal value. According to the spec, it's
            # not valid here anyway (should be in SubSecTimeOriginal), but it does exist sometimes.
            val = val.split('.', 1)[0]
            e['exif_datetime'] = time.mktime(time.strptime(val, '%Y:%m:%d %H:%M:%S'))
        else:
            e['exif_datetime'] = os.path.getmtime(file_path_name)

        gps = exifs.get_location()
        lat = gps['latitude']
        lon = gps['longitude']
        e['latitude'] = round(lat, 4) if lat is not None else lat #TODO sqlite requires (None,) to insert NULL
        e['longitude'] = round(lon, 4) if lon is not None else lon

        #IPTC
        e['tags'] = exifs.get_exif('IPTC Keywords')
        e['title'] = exifs.get_exif('IPTC Object Name')
        e['caption'] = exifs.get_exif('IPTC Caption/Abstract')


        return e


# If being executed (instead of imported), kick it off...
if __name__ == "__main__":
    cache = ImageCache(picture_dir='/home/pi/Pictures', db_file='/home/pi/db.db3', geo_reverse=None)
    #cache.update_cache()
    # items = cache.query_cache("make like '%google%'", "exif_datetime asc")
    #info = cache.get_file_info(12)


import logging
import sys
import argparse
import os
import ssl
import locale
from distutils.dir_util import copy_tree

from picframe import model, viewer_display, controller, interface_kbd, interface_http, __version__

PICFRAME_DATA_DIR = 'picframe_data'

def copy_files(pkgdir, dest, target):
    try:
        fullpath = os.path.join(pkgdir,  target)
        destination = os.path.join(dest,  PICFRAME_DATA_DIR)
        destination = os.path.join(destination,  target)
        copy_tree(fullpath,  destination)
    except:
        raise

def create_config(root):
    fullpath_root = os.path.join(root,  PICFRAME_DATA_DIR)
    fullpath = os.path.join(fullpath_root, 'config')
    source = os.path.join(fullpath, 'configuration_example.yaml')
    destination = os.path.join(fullpath, 'configuration.yaml')
    run_start = os.path.join(fullpath_root, 'run_start.py') # TODO for work-around on RPi4

    try:
        with open (source, "r") as file:
            filedata = file.read()

        print("This will configure ", destination)
        print("To keep default, just hit enter")

        # replace all paths with selected picframe_data path
        filedata = filedata.replace("~/picframe_data", fullpath_root)

        #pic_dir
        pic_dir= input("Enter picture directory [~/Pictures]: ")
        if pic_dir == "":
            pic_dir = "~/Pictures" # convert to absolute path too for work-around on RPi4 running as root
        pic_dir = os.path.expanduser(pic_dir)
        filedata = filedata.replace("~/Pictures", pic_dir)

        #deleted_pictures
        deleted_pictures = input("Enter picture directory [~/DeletedPictures]: ")
        if deleted_pictures == "":
            deleted_pictures = "~/DeletedPictures"
        deleted_pictures = os.path.expanduser(deleted_pictures)
        filedata = filedata.replace("~/DeletedPictures", deleted_pictures)

        #locale
        lan, enc = locale.getlocale()
        if not lan:
            (lan, enc) = ("en_US", "utf8")
        param = input("Enter locale [" + lan + "." + enc + "]: ") or (lan + "." + enc)
        filedata = filedata.replace("en_US.utf8", param)

        with open (destination, "w") as file:
            file.write(filedata)

        with open (run_start, "w") as file: # TODO work-around for RPi4
            file.write("from picframe import start\nstart.main()\n")
    except:
        raise


def check_packages (packages):
    for package in packages:
        try:
            if package == 'paho.mqtt':
                import paho.mqtt
                print(package, ': ',paho.mqtt.__version__)
            elif package == 'ninepatch':
                import ninepatch
                print(package, ': installed, but no version info')
            else:
                print(package, ': ',__import__(package).__version__)
        except ImportError:
            print(package, ': Not found!')

def main():
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    logger = logging.getLogger("start.py")
    logger.info('starting %s', sys.argv)

    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-i", "--initialize",
                        help="creates standard file structure for picframe in destination directory",
                        metavar=('DESTINATION_DIRECTORY'))
    group.add_argument("-v", "--version", help="print version information",
                        action="store_true")
    group.add_argument("configfile", nargs='?', help="/path/to/configuration.yaml")
    args = parser.parse_args()
    if args.initialize:
        if os.geteuid() == 0:
            print("Don't run the initialize step with sudo. It might put the files in the wrong place!")
            return
        pkgdir = sys.modules['picframe'].__path__[0]
        try: 
            dest = os.path.abspath(os.path.expanduser(args.initialize))
            copy_files(pkgdir, dest, 'html')
            copy_files(pkgdir, dest, 'config')
            copy_files(pkgdir, dest, 'data')
            create_config(dest)
            print('created {}/picframe_data'.format(dest))
        except Exception as e:
            print("Can't copy files to: ", args.initialize, ". Reason: ", e)
        return
    elif args.version:
        print("picframe version: ", __version__)
        print("\nChecking required packages......")
        required_packages=[
            'PIL',
            'exifread',
            'pi3d',
            'yaml',
            'paho.mqtt',
            'iptcinfo3',
            'numpy',
            'ninepatch'
        ]
        check_packages(required_packages)
        print("\nChecking optional packages......")
        check_packages(['pyheif'])
        return
    elif args.configfile:
        m = model.Model(args.configfile)
    else:
        m = model.Model()

    v = viewer_display.ViewerDisplay(m.get_viewer_config())
    c = controller.Controller(m, v)
    c.start()

    if m.get_model_config()['use_kbd']:
        interface_kbd.InterfaceKbd(c) # TODO make kbd failsafe

    mqtt_config = m.get_mqtt_config()
    if mqtt_config['use_mqtt']:
        from picframe import interface_mqtt
        mqtt = interface_mqtt.InterfaceMQTT(c, mqtt_config)
        mqtt.start()

    http_config = m.get_http_config()
    model_config = m.get_model_config()
    if http_config['use_http']:
        server = interface_http.InterfaceHttp(c, http_config['path'], model_config['pic_dir'], model_config['no_files_img'], http_config['port'])
        if http_config['use_ssl']:
            server.socket = ssl.wrap_socket(
                server.socket,
                keyfile = http_config['keyfile'],
                certfile = http_config['certfile'],
                server_side=True)
    c.loop()
    if mqtt_config['use_mqtt']:
        mqtt.stop()
    if http_config['use_http']: #TODO objects living in multiple threads issue at shutdown!
        server.stop()
    c.stop()


if __name__=="__main__": 
    main() 



import logging
import os

from  picframe import viewer_display, model

logger = logging.getLogger("test_viewer_display")
logger.setLevel(logging.DEBUG)

def test_model_init():
    m = model.Model('/home/pi/dev/picture_frame/config/configuration.yaml')
    viewer_config = m.get_viewer_config()
    model_config = m.get_model_config()
    viewer = viewer_display.ViewerDisplay(viewer_config)
    viewer.slideshow_start()
    viewer.slideshow_is_running(os.path.expanduser(model_config['no_files_img']))
    assert True == True